from mininet.link import TCLink
from ipmininet.link import IPIntf, TCIntf


class TerraIntf(TCIntf):
    def __init__(self, bwParamMax=10000, *args, **kwargs):
        self.bwParamMax = bwParamMax
        super(TerraIntf, self).__init__(*args, **kwargs)


class TerraLink(TCLink):
    def __init__(self,
                 node1,
                 node2,
                 intf=TerraIntf,
                 cls1=TerraIntf,
                 cls2=TerraIntf,
                 *args, **kwargs):
        super(TerraLink, self).__init__(node1, node2,
                                        intf=intf,
                                        cls1=cls1,
                                        cls2=cls2,
                                        *args, **kwargs)


class TerragraphLink(TerraLink):
    def __init__(self,
                 node1,
                 node2,
                 bw=3600,
                 params1={},
                 params2={},
                 *args, **kwargs):
        #params1.update({"bw": bw, "use_tbf": True})
        #params2.update({"bw": bw, "use_tbf": True})
        super(TerragraphLink, self).__init__(node1, node2,
                                             params1=params1,
                                             params2=params2,
                                             *args, **kwargs)


class WifiLink(TerraLink):
    def __init__(self,
                 node1,
                 node2,
                 *args, **kwargs):
        super(WifiLink, self).__init__(node1, node2,
                                       *args, **kwargs)



import sys
import os
import subprocess
from whichcraft import which


def run_komondor_worker(cfg_file,
                        output_dir=None,
                        komondor_executable=None,
                        komondor_args={}):
    file_name = os.path.basename(cfg_file)
    result_file = os.path.join(output_dir, file_name)
    args = komondor_args.copy()
    args["stats"] = result_file
    k = Komondor(executable=komondor_executable)
    (proc, stdout, stderr) = k.run(cfg_file, **args)
    return stderr


class Komondor(object):
    """Wrapper for Komondor simulator"""
    def __init__(self, executable=None):
        if not executable:
            executable = which('komondor_main')
        self.executable = executable

        if not (self.executable and os.path.isfile(self.executable)):
            raise ValueError("Komondor executable {} not found."
                             .format(executable))

    def run(self, cfg, time=100, seed=1, stats=None, **kwargs):
        '''
        :param int time: Simulation time in seconds.
        :param int seed: Random seed.
        '''

        defaults = {
            "stdout": subprocess.PIPE,
            "stderr": subprocess.PIPE
        }
        defaults.update(kwargs)
        kwargs = defaults

        if not os.path.isfile(cfg):
            raise ValueError("Parameter cfg must be a valid file path.")

        if not (stats):
            stats = "{0}_{2}{1}"\
                    .format(*os.path.splitext(cfg) + ("_result",))

        args = ["--time", str(time),
                "--seed", str(seed),
                "--stats", stats,
                cfg]
        cmd = [self.executable] + args

        proc = subprocess.Popen(cmd, **kwargs)
        (stdout, stderr) = proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError("""Komondor exited with error code {}: \n
                                  stdout: {}\n
                                  stderr: {}\n"""
                               .format(proc.returncode, stdout, stderr))
        return (proc, stdout, stderr)



from mininet.log import info, warn
from mininet.node import OVSSwitch, Node

from .router import TerranetRouter
from .wifi import WifiStation, ConfigurableWifiAccessPoint


class ClientNode(WifiStation):
    def __init__(self, name,
                 *args, **kwargs):
        super().__init__(name,
                         *args, **kwargs)


class DistributionNode(Node):
    pass


class DistributionNode60(TerranetRouter, DistributionNode):
    def __init__(self, name,
                 *args, **kwargs):
        super().__init__(name,
                         *args, **kwargs)


class DistributionNode5_60(ConfigurableWifiAccessPoint, DistributionNode):
    def __init__(self, name, ssid,
                 *args, **kwargs):
        super().__init__(name, ssid,
                         *args, **kwargs)


class Gateway(OVSSwitch):
    def __init__(self, name,
                 **params):
        super().__init__(name,
                         **params)



import os
import signal

from mininet.log import info, warn
from ipmininet.host import IPHost

class IperfHost(IPHost):
    def __init__(self, name,
                 autostart=True,
                 autostart_params=None,
                 *args, **kwargs):
        super().__init__(name,
                         *args, **kwargs)
        if "logfile" in kwargs:
            self.logfile = params["logfile"]
        else:
            self.logfile = "/var/log/iperf_{}.log".format(self.name)
        self.pids = {}
        self.bind_address = None
        self.autostart = autostart
        self.autostart_params = autostart_params

    def run_iperf(self,
                  bin="iperf3",
                  iperf_args="",
                  bind_address=None,
                  *args, **kwargs):
        iface = self.intfList()[0]
        if not bind_address:
            self.bind_address = iface.ip6

        # Kill current processes
        self.stop_iperf()
        # Implement iperf command in subclass
        pass

    def stop_iperf(self):
        for process, pid in self.pids.items():
            try:
                os.killpg(pid, signal.SIGHUP)
                info("Stopped process {} with PID {}.". format(process,
                                                              pid))
            except Exception as e:
                warn("""Could not stop process {} with PID: {}\n
                        {}""".format(process,
                                     pid,
                                     e))
        self.pids.pop('iperf', None)

    def toggle_iperf(self, *args, **kwargs):
        if self.pids.get('iperf'):
            self.stop_iperf()
        else:
            self.run_iperf(*args, **kwargs)
        return self.pids.get('iperf')

    def terminate(self):
        self.stop_iperf()
        super().terminate()


class IperfClient(IperfHost):
    def __init__(self, name,
                 host=None,
                 netstats_log=None,
                 *args, **kwargs):
        self.host = host
        self.netstats_log = netstats_log
        super().__init__(name,
                         *args, **kwargs)

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, host):
        if isinstance(host, IperfServer):
            self._host = host.intfList()[0].ip6
        else:
            self._host = host

    def run_iperf(self,
                  bin="iperf3",
                  iperf_args="-6 -t0 -i10 -l1400 -w65M -Z",
                  *args, **kwargs):
        super().run_iperf(bin=bin,
                          iperf_args=iperf_args,
                          *args, **kwargs)

        if not self.host:
            raise ValueError("""Host attribute must be set before running
                                iperf client.""")

        iface = self.intfList()[0]
        if not self.netstats_log:
            self.netstats_log = "/var/log/{}_{}_netstats.log".format(iface.name,
                                                                   iface.ip6)
        netstats_cmd = ("(while :; do "
                       "cat /proc/net/dev | grep {intf} > {logfile} 2>&1; "
                       "sleep {interval}; done) &").format(intf=iface.name,
                                                           logfile=self.netstats_log,
                                                           interval=2)
        p_netstats = self.popen(netstats_cmd, shell=True)
        self.pids.update({"netstats_logger": p_netstats.pid})

        # --logfile option requires iperf3 >= 3.1
        cmd = ("until ping6 -c1 {host} >/dev/null 2>&1; do :; done; "
               "{bin} {iperf_args} "
               "-c {host} "
               "-B {bind} "
               "--logfile {log}").format(bin=bin,
                                         iperf_args=iperf_args,
                                         host=self.host,
                                         bind=self.bind_address,
                                         log=self.logfile)
        p = self.popen(cmd, shell=True)
        self.pids.update({"iperf": p.pid})
        return p

    def stop_iperf(self):
        super().stop_iperf()
        if self.netstats_log:
            os.remove(self.netstats_log)
            self.netstats_log = None
        self.pids.pop('netstats_logger', None)


class IperfReverseClient(IperfClient):
    def __init__(self, name,
                 *args, **kwargs):
        super().__init__(name,
                         *args, **kwargs)

    def run_iperf(self,
                  bin="iperf3",
                  iperf_args="-6 -R -t0 -i10 -l1400 -w 65M -Z",
                  *args, **kwargs):
        super().run_iperf(bin=bin,
                          iperf_args=iperf_args,
                          *args, **kwargs)


class IperfServer(IperfHost):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name,
                         *args, **kwargs)

    def run_iperf(self,
                  bin="iperf3",
                  iperf_args="-s",
                  *args, **kwargs):
        super().run_iperf(bin=bin,
                          iperf_args=iperf_args,
                          *args, **kwargs)

        # --logfile option requires iperf3 >= 3.1
        cmd = ("{bin} {iperf_args} "
               "-B {bind} "
               "--logfile {log}").format(bin=bin,
                                         iperf_args=iperf_args,
                                         bind=self.bind_address,
                                         log=self.logfile)
        p = self.popen(cmd, shell=True)
        self.pids.update({"iperf": p.pid})
        return p



from ryu.controller import ofp_event
from ryu.controller.handler import CONFIG_DISPATCHER, MAIN_DISPATCHER
from ryu.controller.handler import set_ev_cls
from ryu.controller.event import EventBase
from ryu.ofproto import ofproto_v1_3
from ryu.lib.packet import packet
from ryu.lib.packet import ethernet
from ryu.lib.packet import ipv6

from .mac_learning_pipeline import MacLearningPipeline
from .ipv6_address_helper import IPv6AddressHelper


class EventCustomerFlowAdded(EventBase):
    def __init__(self, customer_ipv6):
        super().__init__()
        self.customer_ipv6 = customer_ipv6


class EventCustomerFlowRemoved(EventBase):
    def __init__(self, customer_ipv6):
        super().__init__()
        self.customer_ipv6 = customer_ipv6


class CustomerFlowPipeline(MacLearningPipeline):
    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]
    SRC_MAC_TABLE = 100
    DST_MAC_TABLE = 101
    DN_TABLE = 3
    FIRST_CUSTOMER_TABLE = 4
    NEXT_TABLE = FIRST_CUSTOMER_TABLE
    CUSTOMER_FLOW_IDLE_TIMEOUT = 10

    _EVENTS = [
        EventCustomerFlowAdded,
        EventCustomerFlowRemoved
    ]

    def __init__(self, *args, **kwargs):
        self.customer_table_allocation = {}
        super().__init__(*args, **kwargs)

    def _notify_customer_event_flow_added(self, customer_ipv6):
        ev = EventCustomerFlowAdded(customer_ipv6)
        self.send_event_to_observers(ev)
        self.logger.info('CustomerFlowPipeline: '
                         'Sent event {} to observers.'
                         .format(ev.__class__.__name__))

    def _notify_customer_event_flow_removed(self, customer_ipv6):
        ev = EventCustomerFlowRemoved(customer_ipv6)
        self.send_event_to_observers(ev)
        self.logger.info('CustomerFlowPipeline: '
                         'Sent event {} to observers.'
                         .format(ev.__class__.__name__))

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)
    def switch_features_handler(self, ev):
        super().switch_features_handler(ev)
        datapath = ev.msg.datapath
        self._init_table_zero(datapath)
        self._init_dn_table(datapath)

    def _init_table_zero(self, datapath):
        ofp_parser = datapath.ofproto_parser
        table_id = 0

        # Send traffic to/ from DN subnets to DN table
        dn_table_id = self.__class__.DN_TABLE
        dn_match_inst = [ofp_parser.OFPInstructionGotoTable(dn_table_id)]

        dst_match = ofp_parser.OFPMatch(
            eth_type=0x86dd,
            ipv6_dst=("fd00:0:0:8000::", "ffff:ffff:ffff:8000::"))
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    table_id=table_id,
                                    match=dst_match,
                                    instructions=dn_match_inst)
        datapath.send_msg(mod)

        src_match = ofp_parser.OFPMatch(
            eth_type=0x86dd,
            ipv6_src=("fd00:0:0:8000::", "ffff:ffff:ffff:8000::"))
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    table_id=table_id,
                                    match=src_match,
                                    instructions=dn_match_inst)
        datapath.send_msg(mod)

        self.logger.info("CustomerFlowPipeline: Initialized table {}. "
                         "Processing dn subnet match at table {}.".format(
                             table_id,
                             dn_table_id))

        # Send other traffic directly to MAC learning table
        src_mac_table_id = self.__class__.SRC_MAC_TABLE
        table_miss_inst = [ofp_parser.OFPInstructionGotoTable(src_mac_table_id)]
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    table_id=table_id,
                                    match=ofp_parser.OFPMatch(),
                                    instructions=table_miss_inst)
        datapath.send_msg(mod)
        self.logger.info("CustomerFlowPipeline: Initialized table {}. "
                         "On table-miss send to table {}.".format(
                             table_id,
                             src_mac_table_id))

    def _init_dn_table(self, datapath):
        ofproto = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        table_id = self.__class__.DN_TABLE
        actions = [ofp_parser.OFPActionOutput(ofproto.OFPP_CONTROLLER)]
        inst = [ofp_parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS,
                                                 actions)]
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    table_id=table_id,
                                    priority=0,
                                    match=ofp_parser.OFPMatch(),
                                    instructions=inst)
        datapath.send_msg(mod)
        self.logger.info("CustomerFlowPipeline: Initialized DN table {}. "
                         "On table-miss send to controller.".format(table_id))

    @set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)
    def _packet_in_handler(self, ev):
        # MAC learning
        super()._packet_in_handler(ev)
        msg = ev.msg
        in_port = msg.match["in_port"]

        table_id = msg.table_id
        self.logger.info("CustomerFlowPipeline: Packet in at port {} "
                         "from table {}.".format(in_port,
                                                 table_id))

        if table_id == self.__class__.DN_TABLE:
            self._handle_new_distribution_subnet(ev)
        if (table_id >= self.__class__.FIRST_CUSTOMER_TABLE and
                table_id < self.__class__.SRC_MAC_TABLE):
            self._handle_new_customer_address(ev)

    def _handle_new_distribution_subnet(self, ev):
        msg = ev.msg
        datapath = msg.datapath
        ofp_parser = datapath.ofproto_parser
        buffer_id = msg.buffer_id
        pkt = packet.Packet(msg.data)

        pkt_ipv6 = pkt.get_protocols(ipv6.ipv6)[0]
        subnet_prefix = IPv6AddressHelper.subnet_prefix(pkt_ipv6.src)
        subnet_id = IPv6AddressHelper.subnet_id(pkt_ipv6.src)
        dn_id = IPv6AddressHelper.distribution_id(pkt_ipv6.src)
        cn_id = IPv6AddressHelper.client_id(pkt_ipv6.src)
        self.logger.info("CustomerFlowPipeline: Distribution/ Client network "
                         "detected. IPv6: {}, "
                         "Subnet prefix: {}, "
                         "Subnet id: {}, "
                         "Distibution node id: {}, "
                         "Client node id: {}.".format(pkt_ipv6,
                                                      subnet_prefix,
                                                      subnet_id,
                                                      dn_id,
                                                      cn_id))

        # Process traffic to/ from customers at customer table
        table_id = self.__class__.DN_TABLE
        dn_table_id = self._get_customer_table_id(ev)
        customer_inst = [ofp_parser.OFPInstructionGotoTable(dn_table_id)]

        dst_customer_match = ofp_parser.OFPMatch(
            eth_type=0x86dd,
            ipv6_dst=("{}:8000::".format(subnet_prefix),
                      "ffff:ffff:ffff:ffff:8000::"))
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    priority=200,
                                    table_id=table_id,
                                    match=dst_customer_match,
                                    instructions=customer_inst)
        datapath.send_msg(mod)

        src_customer_match = ofp_parser.OFPMatch(
            eth_type=0x86dd,
            ipv6_src=("{}:8000::".format(subnet_prefix),
                      "ffff:ffff:ffff:ffff:8000::"))
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    priority=200,
                                    table_id=table_id,
                                    match=src_customer_match,
                                    instructions=customer_inst)
        datapath.send_msg(mod)

        # Process other traffic from subnet at MAC learning table
        src_mac_table_id = self.__class__.SRC_MAC_TABLE
        client_net_inst = [ofp_parser.OFPInstructionGotoTable(src_mac_table_id)]

        dst_client_net_match = ofp_parser.OFPMatch(
            eth_type=0x86dd,
            ipv6_dst=("{}::".format(subnet_prefix),
                      "ffff:ffff:ffff:ffff::"))
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    priority=100,
                                    table_id=table_id,
                                    match=dst_client_net_match,
                                    instructions=client_net_inst)
        datapath.send_msg(mod)

        src_client_net_match = ofp_parser.OFPMatch(
            eth_type=0x86dd,
            ipv6_src=("{}::".format(subnet_prefix),
                      "ffff:ffff:ffff:ffff::"))
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    buffer_id=buffer_id,
                                    priority=100,
                                    table_id=table_id,
                                    match=src_client_net_match,
                                    instructions=client_net_inst)
        datapath.send_msg(mod)

    def _get_customer_table_id(self, ev):
        msg = ev.msg
        datapath = msg.datapath
        pkt = packet.Packet(msg.data)

        pkt_ipv6 = pkt.get_protocols(ipv6.ipv6)[0]
        dn_id = IPv6AddressHelper.distribution_id(pkt_ipv6.src)
        distribution_net = IPv6AddressHelper.distribution_net(pkt_ipv6.src)

        if not self.customer_table_allocation.get(dn_id):
            table_id = self.__class__.NEXT_TABLE
            self.customer_table_allocation[dn_id] = { "table_id": table_id }
            if distribution_net:
                dn_entry = self.customer_table_allocation[dn_id]
                dn_entry["distribution_net"] = distribution_net
            self.__class__.NEXT_TABLE += 1
            self._init_customer_table(datapath, table_id)

            self.logger.info("CustomerFlowPipeline: "
                             "No table allocated for DN yet. "
                             "Allocated table {}".format(table_id))

        return self.customer_table_allocation[dn_id]["table_id"]

    def _init_customer_table(self, datapath, table_id):
        ofproto = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        actions = [ofp_parser.OFPActionOutput(ofproto.OFPP_CONTROLLER)]
        inst = [ofp_parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS,
                                                 actions)]
        mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                    table_id=table_id,
                                    priority=0,
                                    match=ofp_parser.OFPMatch(),
                                    instructions=inst)
        datapath.send_msg(mod)
        self.logger.info("CustomerFlowPipeline: Initialized table {}. "
                         "On table-miss send to controller.".format(table_id))

    def _handle_new_customer_address(self, ev):
        msg = ev.msg
        table_id = msg.table_id
        datapath = msg.datapath
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser
        buffer_id = msg.buffer_id
        pkt = packet.Packet(msg.data)
        pkt_ipv6 = pkt.get_protocols(ipv6.ipv6)[0]
        idle_timeout = self.__class__.CUSTOMER_FLOW_IDLE_TIMEOUT

        src_mac_table_id = self.__class__.SRC_MAC_TABLE
        inst = [ofp_parser.OFPInstructionGotoTable(src_mac_table_id)]

        customer_ips = []
        if IPv6AddressHelper.is_customer_address(pkt_ipv6.src):
            customer_ips.append(pkt_ipv6.src)
        if IPv6AddressHelper.is_customer_address(pkt_ipv6.dst):
            customer_ips.append(pkt_ipv6.dst)
        if not customer_ips:
            raise RuntimeError("CustomerFlowPipeline: Packet does not include "
                               "customer ipv6 address.")

        for customer_ip in customer_ips:
            dst_match = ofp_parser.OFPMatch(
                eth_type=0x86dd,
                ipv6_dst=customer_ip)
            mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                        table_id=table_id,
                                        match=dst_match,
                                        idle_timeout=idle_timeout,
                                        instructions=inst,
                                        flags=ofp.OFPFF_SEND_FLOW_REM)
            datapath.send_msg(mod)

            src_match = ofp_parser.OFPMatch(
                eth_type=0x86dd,
                ipv6_src=customer_ip)
            mod = ofp_parser.OFPFlowMod(datapath=datapath,
                                        buffer_id=buffer_id,
                                        table_id=table_id,
                                        match=src_match,
                                        instructions=inst)
            datapath.send_msg(mod)

            self._notify_customer_event_flow_added(customer_ip)
            self.logger.info("CustomerFlowPipeline: "
                             "New customer: {} "
                             "added to table {}".format(pkt_ipv6,
                                                        table_id))

    @set_ev_cls(ofp_event.EventOFPFlowRemoved, MAIN_DISPATCHER)
    def _flow_removed_handler(self, ev):
        msg = ev.msg
        datapath = msg.datapath
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        reason = 'unknown'
        if msg.reason == ofp.OFPRR_IDLE_TIMEOUT:
            reason = 'IDLE TIMEOUT'
        elif msg.reason == ofp.OFPRR_HARD_TIMEOUT:
            reason = 'HARD TIMEOUT'
        elif msg.reason == ofp.OFPRR_DELETE:
            reason = 'DELETE'
        elif msg.reason == ofp.OFPRR_GROUP_DELETE:
            reason = 'GROUP DELETE'

        table_id = msg.table_id
        ipv6_dst = msg.match['ipv6_dst']

        self.logger.info('CustomerFlowPipeline: '
                         'A flow was removed from table {}. '
                         'Reason: {}.'
                         .format(table_id, reason))

        if ipv6_dst:
            src_match = ofp_parser.OFPMatch(
                eth_type=0x86dd,
                ipv6_src=ipv6_dst)
            mod = ofp_parser.OFPFlowMod(command=ofp.OFPFC_DELETE,
                                        datapath=datapath,
                                        table_id=table_id,
                                        match=src_match)
            datapath.send_msg(mod)
            self.logger.info('CustomerFlowPipeline: '
                             'Match for ipv6_dst {} at table {} timed out. '
                             'Removed match for ipv6_src.'
                             .format(ipv6_dst, table_id))
        self._notify_customer_event_flow_removed(ipv6_dst)



from ipmininet.router import OpenrRouter
from ipmininet.router.config import OpenrRouterConfig

class TerranetRouter(OpenrRouter):
    def __init__(self, name,
                 config=OpenrRouterConfig,
                 lo_addresses=(),
                 *args, **kwargs):
        super().__init__(name,
                         config=config,
                         lo_addresses=lo_addresses,
                         *args, **kwargs)



import os
import proxy
from multiprocessing import Process
import shlex, subprocess

class NamespaceProxy:
    def __init__(self, nspid, port,
                 num_workers=1,
                 hostname='::',
                 *args, **kwargs):
        self.nspid = nspid
        self.port = port
        self.num_workers = num_workers
        self.hostname = hostname
        self.process = None

    def start_proxy_process(self):
        args = (self.nspid, self.port,)
        kwargs = {'num_workers': self.num_workers,
                  'hostname': self.hostname}
        env = os.environ.copy()
        env['NSPID'] = str(self.nspid)
        cmd = 'proxy' \
              ' --num-workers {workers}' \
              ' --hostname {host}' \
              ' --port {port}' \
              ' --plugins terranet.proxy.NamespaceProxyPlugin' \
              ' --enable-web-server'.format(workers=self.num_workers,
                                            host=self.hostname,
                                            port=self.port)
        args = shlex.split(cmd)
        p = subprocess.Popen(args, env=env, stdin=subprocess.PIPE,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE)
        self.process = p
        return p



from mininet.log import setLogLevel, info, debug
from ipmininet.cli import IPCLI

from terranet.topo import HybridBackupTerragraphTopo
from terranet.net import Terranet
from terranet.controller import RyuManager


if __name__ == '__main__':
    setLogLevel('info')
    topo = HybridBackupTerragraphTopo()
    net = Terranet(topo=topo)
    module_path = 'terranet.ryu.app'
    apps = ('customer_flow_pipeline',
            'customer_stats_monitor')
    modules = [ '.'.join([module_path, x]) for x in apps ]
    ryu_manager = RyuManager('ryu-manager', modules=modules)
    net.addController(ryu_manager)
    net.start()
    IPCLI(net)
    net.stop()



import os
import collections
import math
import itertools
import multiprocessing
from functools import partial
from configparser import ConfigParser

from mininet.log import info, warn

from ..event import KomondorConfigChangeEvent, ChannelSwitchEvent, \
    Sub6GhzEmulatorRegistrationEvent, \
    Sub6GhzEmulatorCancelRegistrationEvent
from .komondor import run_komondor_worker
from .komondor_config import KomondorConfig, KomondorSystemConfig, \
                             read_komondor_configs, read_komondor_results


class Sub6GhzEmulator(object):
    def __init__(self,
                 net=None,
                 komondor_executable=None,
                 komondor_config_dir=None,
                 komondor_system_cfg=None,
                 komondor_configs=None,
                 komondor_results=None,
                 current_komondor_file=None,
                 current_komondor_config=None,
                 current_komondor_result=None):
        self.net = net
        self.komondor_executable = komondor_executable
        self.komondor_config_dir = os.path.abspath(komondor_config_dir)
        self.komondor_input_dir = f'{self.komondor_config_dir}/input'
        self.komondor_output_dir = f'{self.komondor_config_dir}/output'
        if not komondor_system_cfg:
            komondor_system_cfg = KomondorSystemConfig()
        self.komondor_system_cfg = komondor_system_cfg
        self.komondor_configs = komondor_configs
        self.komondor_results = komondor_results
        self.current_komondor_file = current_komondor_file
        self.current_komondor_config = current_komondor_config
        self.current_komondor_result = current_komondor_result

    def update(self, evt):
        if isinstance(evt, KomondorConfigChangeEvent):
            self.handle_komondor_config_change(evt)
        elif isinstance(evt, ChannelSwitchEvent):
            self.handle_channel_switch(evt)
        elif isinstance(evt, Sub6GhzEmulatorRegistrationEvent):
            self.handle_registration(evt)
        elif isinstance(evt, Sub6GhzEmulatorCancelRegistrationEvent):
            self.handle_cancel_registration(evt)

    def handle_komondor_config_change(self, evt):
        info("Sub6GhzEmulator: Changing config of node {}.\n"
             .format(evt.node.name))
        info("Sub6GhzEmulator: New config {}\n".format(evt.update))
        evt.result = True
        evt.message = "OK\n"
        evt.set()

    def handle_channel_switch(self, evt):
        info("Sub6GhzEmulator: Node {} triggered channel switch.\n"
             .format(evt.node.name))
        ap = evt.node
        client_nodes = ap.connected_stations()
        try:
            self.adjust_station_wifi_config(ap)
            self.apply_wifi_config()
            evt.result = True
            evt.message = "New config file: {}\n"\
                          .format(self.current_komondor_file)
            evt.set()
        except RuntimeError as err:
            error_message = """Sub6GhzEmulator: Error {}
                               Rolling back to previos config: {}.\n"""\
                            .format(err, self.current_komondor_file)
            warn(error_message)
            for node in (client_nodes + [ap]):
                node.update_komondor_config(evt.old_channel_config)
            self.apply_wifi_config()
            evt.result = False
            evt.message = error_message
            evt.set()

    def handle_registration(self, evt):
        info("Sub6GhzEmulator: Registering node {}\n".format(evt.node.name))
        evt.result = True
        evt.message = "OK\n"
        evt.set()

    def handle_cancel_registration(self, evt):
        info("Sub6GhzEmulator: Unregistering node {}\n"
             .format(evt.node.name))
        evt.result = True
        evt.message = "OK\n"
        evt.set()

    def apply_wifi_config(self):
        info("Sub6GhzEmulator: Trying to apply new network config.\n")
        new_config = self.wifi_config()
        config_tuple = self.find_kommondor_config(new_config)
        if not config_tuple:
            warn("No config found for new network configs.\n")
            raise RuntimeError("Current config not found.")
        else:
            file_name = config_tuple[0]
            config = config_tuple[1]

        self.current_komondor_file = file_name
        self.current_komondor_config = self.komondor_configs[file_name]
        self.current_komondor_result = self.komondor_results[file_name]
        info('Sub6GhzEmulator: Trying to apply new config file {}.\n'
             .format(file_name))
        self.apply_results()
        info("Sub6GhzEmulator: Network config {} successfully applied.\n"
             .format(config.cfg_file))

    def apply_results(self):
        for ap in self.net.access_points():
            stas = ap.connected_stations()
            for sta in stas:
                komondor_name = sta.komondor_config.name
                result = self.current_komondor_result[komondor_name]
                links = self.net.linksBetween(ap, sta)
                for link in links:
                    bw = int(result.getint("throughput") / 1000000)
                    # Dirty fix for ZeroDivisionError in mininet
                    if bw == 0:
                        bw = 1
                    delay_value = result.get('delay')
                    if delay_value in ['nan', '-nan']:
                        delay_value = '0'
                    delay = '{}ms'.format(round(float(delay_value)))
                    link.intf1.config(bw=bw, delay=delay, use_tbf=True)
                    link.intf2.config(bw=bw, delay=delay, use_tbf=True)

    def wifi_config(self):
        config = KomondorConfig()
        config_dict = collections.OrderedDict()
        config_dict["System"] = collections.OrderedDict(
            **self.komondor_system_cfg)
        node_configs = [node.komondor_config for node in self.net.wifi_nodes()]

        for cfg in node_configs:
            config_dict[cfg.name] = collections.OrderedDict(**cfg)

        config.read_dict(config_dict)
        return config

    def adjust_station_wifi_config(self, ap):
        wlan_code = ap.komondor_config["wlan_code"]
        channel_params = ap.channel_config()
        for sta in ap.connected_stations():
            sta.komondor_config.update({"wlan_code": wlan_code})
            sta.komondor_config.update(channel_params)
        return channel_params

    def build_komondor(self, use_cache=True):
        configs = []
        access_points = self.net.access_points()

        # No simulation needed
        if not access_points:
            return None

        for ap in access_points:
            self.adjust_station_wifi_config(ap)

        channels = sorted([ ap.available_channels for ap in access_points ])
        channel_combinations = list(itertools.product(*channels))

        for channels in channel_combinations:
            config = self.__build_channel_config(channels)
            configs.append(config)

        cached_config_dict = read_komondor_configs(self.komondor_input_dir)
        cached_configs = list(cached_config_dict.values())
        if not use_cache or configs != cached_configs:
            info("Not using cached komondor config.\n")
            self.delete_cache()
            self.write_komondor_configs(configs)
            self.komondor_configs = read_komondor_configs(
                self.komondor_input_dir)
            info("Simulating komondor configurations. "
                 "This might take a while.\n")
            import time
            start = time.time()
            self.presimulate()
            end = time.time()
            time_elapsed = end - start
            info("Simulations finished."
                 "Time elapsed: {} seconds.\n".format(time_elapsed))
        else:
            info("Using cached komondor config.\n")
            self.komondor_configs = cached_config_dict
        self.komondor_results = read_komondor_results(
            self.komondor_output_dir)
        return self

    def __build_channel_config(self, channels):
        config_dict = collections.OrderedDict()
        config_dict["System"] = collections.OrderedDict(
            **self.komondor_system_cfg)
        for i, ap in enumerate(self.net.access_points()):
            channel = channels[i]
            cfg = self.__build_bss_config(ap, channel)
            config_dict.update(cfg)
        config = KomondorConfig()
        config.read_dict(config_dict)
        return config

    def __build_bss_config(self, ap, channel):
        config = collections.OrderedDict()
        channel_cfg = channel.komondor_channel_params
        primary_channel = channel_cfg["min_channel_allowed"]
        ap_cfg = collections.OrderedDict(**ap.komondor_config)
        ap_cfg.update(channel_cfg)
        ap_cfg.update({"primary_channel": primary_channel})
        ap_name = ap.komondor_config.name
        config[ap_name] = collections.OrderedDict(ap_cfg)
        for sta in ap.connected_stations():
            cfg = collections.OrderedDict(**sta.komondor_config)
            cfg.update({"wlan_code": ap_cfg["wlan_code"],
                        "primary_channel": primary_channel})
            cfg.update(channel_cfg)
            sta_name = sta.komondor_config.name
            config[sta_name] = collections.OrderedDict(cfg)
        return config

    def write_komondor_configs(self, configs):
        config_map = []
        max_digits = int(math.log10(len(configs))) + 1
        for i, config in enumerate(configs):
            suffix = format(i, "0{}".format(max_digits))
            fname = "{}.cfg".format(suffix)
            fpath = os.path.join(self.komondor_input_dir, fname)
            with open(fpath, 'w') as f:
                config.write(f)
            config_map.append((fname, config))
        return config_map

    def presimulate(self):
        cfg_files = [os.path.join(self.komondor_input_dir, f)
                     for f in os.listdir(self.komondor_input_dir)
                     if f.endswith(".cfg")]
        pool = multiprocessing.Pool()

        f = partial(run_komondor_worker,
                    output_dir=self.komondor_output_dir)
        pool.map(f, cfg_files)

    def delete_cache(self):
        for d in [self.komondor_input_dir, self.komondor_output_dir]:
            for f in os.listdir(d):
                if f.endswith(".cfg"):
                    os.remove(os.path.join(d, f))

    def find_kommondor_config(self, config):
        for fname, cfg in self.komondor_configs.items():
            if cfg == config:
                return (fname, cfg)
        return None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.
    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import proxy
import netns
from uuid import UUID
from typing import List, Tuple
from urllib import parse as urlparse

from proxy.common.flags import Flags
from proxy.core.connection import TcpClientConnection
from proxy.core.event import EventQueue
from proxy.common.constants import DEFAULT_BUFFER_SIZE
from proxy.common.utils import socket_connection
from proxy.http.parser import HttpParser
from proxy.http.websocket import WebsocketFrame
from proxy.http.server import HttpWebServerBasePlugin, httpProtocolTypes


class NamespaceProxyPlugin(HttpWebServerBasePlugin):
    """Extend in-built Web Server to add Reverse Proxy capabilities.
    This example plugin is equivalent to following Nginx configuration:
        location /get {
            proxy_pass http://httpbin.org/get
        }
    Example:
        $ curl http://localhost:9000/get
        {
          "args": {},
          "headers": {
            "Accept": "*/*",
            "Host": "localhost",
            "User-Agent": "curl/7.64.1"
          },
          "origin": "1.2.3.4, 5.6.7.8",
          "url": "https://localhost/get"
        }
    """

    def __init__(
            self,
            uid: UUID,
            flags: Flags,
            client: TcpClientConnection,
            event_queue: EventQueue):
        super().__init__(uid, flags, client, event_queue)
        self.location = str = r'/*'
        self.upstream: Tuple[str, int] = ('localhost', 80)
        try:
            self.nspid = int(os.environ['NSPID'])
        except KeyError as e:
            raise(EnvironmentError('Environment variable NAMESPACE must be set'
                                   'to run NamespaceProxyPlugin'))

    def routes(self) -> List[Tuple[int, str]]:
        return [
            (httpProtocolTypes.HTTP, self.location),
            (httpProtocolTypes.HTTPS, self.location)
        ]

    def handle_request(self, request: HttpParser) -> None:
        nspid = self.nspid
        with netns.NetNS(nspid=nspid):
            with socket_connection(self.upstream) as conn:
                conn.send(request.build())
                data = conn.recv(DEFAULT_BUFFER_SIZE)
                while(data):
                    self.client.queue(memoryview(data))
                    data = conn.recv(DEFAULT_BUFFER_SIZE)

    def on_websocket_open(self) -> None:
        raise NotImplementedError()

    def on_websocket_message(self, frame: WebsocketFrame) -> None:
        raise NotImplementedError()

    def on_websocket_close(self) -> None:
        raise NotImplementedError()



import os
import time
import random
import numpy as np

from mininet.log import info

import terranet.ryu.app as ryu_app_module
from terranet.topo import HybridVirtualFiberTopo
from terranet.net import Terranet
from terranet.controller import RyuManager
from terranet.simulator import TerranetSimulator

def toogle_customer_arrival(net, wait, customer):
    info(f'Next arrival after {wait} seconds.\n')
    time.sleep(wait)
    info(f'Toggling iperf of customer {customer}.\n')
    net[customer].toggle_iperf()


if __name__ == '__main__':
    seed = 2000
    rng = random.Random(seed)
    arrivals = 10
    customers = [f'h{x}{y}' for x in range(2,6) for y in ('a1','b1','c1')]
    customer_events = rng.choices(customers, k=arrivals)
    np.random.seed(seed)
    lam = 90
    waits = np.random.poisson(lam, arrivals)
    topo = HybridVirtualFiberTopo()
    net = Terranet(topo=topo,
                   build=False)
    apps = ('customer_flow_pipeline',
            'customer_stats_monitor',
            'channel_assignment_oracle')
    modules = [ '.'.join(['terranet.ryu.app', x]) for x in apps ]
    config_file =  os.path.join(os.path.dirname(ryu_app_module.__file__),
                                'cfg/channel_oracle.conf')
    ryuArgs = (f'--config-file {config_file}')
    ryu_manager = RyuManager('ryu-manager', ryuArgs, modules=modules)
    net.addController(ryu_manager)
    with TerranetSimulator('channel_oracle_simulation', net, clean_after=False) as net:
        for i in range(0, arrivals):
            toogle_customer_arrival(net, waits[i], customer_events[i])
        time.sleep(90)



from functools import partial

from flask import Flask, request, jsonify, abort
from flask.helpers import locked_cached_property

from .wifi.channel import Channel


class ConfigAPI(Flask):
    def __init__(self,
                 import_name,
                 node,
                 static_url_path=None,
                 static_folder=None,
                 static_host=None,
                 host_matching=False,
                 subdomain_matching=False,
                 template_folder=None,
                 instance_path=None,
                 instance_relative_config=False,
                 root_path=None):

        super(ConfigAPI, self).__init__(
                import_name,
                static_url_path=static_url_path,
                static_folder=static_folder,
                static_host=static_host,
                host_matching=host_matching,
                subdomain_matching=subdomain_matching,
                template_folder=template_folder,
                instance_path=instance_path,
                instance_relative_config=instance_relative_config,
                root_path=root_path
        )

        self.node = node
        self._routes_init()

    def _routes_init(self):
        self.add_url_rule("/config", "get_config", self.get_config,
                          methods=["GET"])

        self.add_url_rule("/channel", "switch_channel", self.switch_channel,
                          methods=["POST"])

    def get_config(self):
        return jsonify(self.node.komondor_config)

    def switch_channel(self):
        payload = request.json
        if not payload:
            # TODO add error message
            abort(400)
        if "channel" in payload:
            channel_num = payload["channel"]
            channel = Channel(channel_num)
            (success, message) = self.node.switch_channel(channel)
            if success:
                return "{}".format(message)
            else:
                # TODO add error message
                abort(400)
        else:
            # TODO add error message
            abort(400)



import ipaddress

from .terratopo import Terratopo
from ..wifi.channel import Channel


class HybridVirtualFiberTopo(Terratopo):
    '''
    mmWave Distribution Network backhaul
    ====================================

            +------+   60GHz  +------+   60GHz  +------+
            | dn_a |- - - - - | dn_b |- - - - - | dn_c |
            +------+          +------+          +------+
                |
                | cable
                |
            + -_-_- +
            |  _X_  | gw
            + - - - +
             | |    |
             | |    |  cable
             .-~~~-.|
     .- ~ ~-(| |     \)_ _
    /        / |      \    ~ -.
   |     s1a1 s1a2 ... s3c3     \ 
    \                        _.~'
      ~- . ______________ . -



    WiFi edge
    =========

        h1a1  ...  h5a1   h1b1  ...   h5b1  h1c1  ...   h5c1    (CPEs)
          \    |    /       \    |    /       \    |    /
       ^   \   |   /         \   |   /         \   |   /      -- cable
       |    +-----+           +-----+           +-----+
    20 |    |cn_a1|           |cn_b1|           |cn_b2|       (STAs/ CNs)
       |    +-----+           +-----+           +-----+
       |       .                 .                 .
       |       .                 .                 .
    10 |       .                 .                 .        ... 5GHz WiFi
       |       .                 .                 .
       |       .                 .                 .
       |    +------+          +------+          +------+
     0 |    | dn_a |          | dn_b |          | dn_c |      (APs/ DNs)
       |    +------+          +------+          +------+
       ----------------------------------------------->
     -15       0       25       50       75        100
    '''

    def build(self, *args, **kwargs):
        channel_list = [Channel(42), Channel(36), Channel(38), Channel(40),
                        Channel(44), Channel(46), Channel(48)]

        # Segment A
        dn_a = self.add_distribution_node_5_60(
            'dn_a',
            ssid='A',
            available_channels=channel_list,
            coordinates={'x': 0, 'y': 0, 'z': 0},
            proxy_port=8199)

        cn_a1 = self.add_client_node('cn_a1',
                                     coordinates={'x': 0, 'y': 20, 'z': 0})
        self.add_wifi_link(dn_a, cn_a1)


        # Segment B
        dn_b = self.add_distribution_node_5_60(
            'dn_b',
            ssid='B',
            available_channels=channel_list,
            coordinates={'x': 20, 'y': 0, 'z': 0},
            proxy_port=8299)

        cn_b1 = self.add_client_node('cn_b1',
                                     coordinates={'x': 50, 'y': 20, 'z': 0})
        self.add_wifi_link(dn_b, cn_b1)


        # Segment C
        dn_c = self.add_distribution_node_5_60(
            'dn_c',
            ssid='C',
            available_channels=channel_list,
            coordinates={'x': 40, 'y': 0, 'z': 0},
            proxy_port=8399)

        cn_c1 = self.add_client_node('cn_c1',
                                     coordinates={'x': 100, 'y': 20, 'z': 0})
        self.add_wifi_link(dn_c, cn_c1)


        # Add WiFi 60GHz Links between DN_A and DN_B
        self.add_terragraph_link(dn_a, dn_b)

        # Add WiFi 60GHz Links between DN_B and DN_C
        self.add_terragraph_link(dn_b, dn_c)

        # Add GW switch behind DN_A
        gw = self.add_gateway('s1')
        self.add_ip_link(gw, dn_a)

        # cn_a1 customer flows
        self._create_customer_flows(cn_a1, gw, 'a1',
                                    subnet=('10.145.128.0/24',
                                            'fd00:0:0:8101:8000::0/80'))

        # cn_b1 customer flows
        self._create_customer_flows(cn_b1, gw, 'b1',
                                    subnet=('10.161.128.0/24',
                                            'fd00:0:0:8201:8000::0/80'))

        # cn_c1 customer flows
        self._create_customer_flows(cn_c1, gw, 'c1',
                                    subnet=('10.177.128.0/24',
                                            'fd00:0:0:8301:8000::0/80'))
        super().build(*args, **kwargs)

    def _create_customer_flows(self, cn, gw, cn_suffix,
                               num_flows=5,
                               num_enabled=1,
                               subnet=('10.128.128.0/24',
                                       'fd00:0:0:8000:8000::0/80')):

        ipv4_net_address, ipv4_prefix_str = subnet[0].split('/')
        ipv4_prefix = int(ipv4_prefix_str)
        ipv6_net_address, ipv6_prefix_str = subnet[1].split('/')
        ipv6_prefix = int(ipv6_prefix_str)
        ipv4_subnets = list(
            ipaddress.ip_network('{}/{}'.format(ipv4_net_address,
                                                ipv4_prefix - 7))
                     .subnets(new_prefix=ipv4_prefix))
        ipv6_subnets = list(
            ipaddress.ip_network('{}/{}'.format(ipv6_net_address,
                                                ipv6_prefix - 7))
                     .subnets(new_prefix=ipv6_prefix))

        for i in range(1, (num_flows+1)):
            suffix = f'{i}{cn_suffix}'
            if i <= num_enabled:
                autostart_client = True
            else:
                autostart_client = False

            self.add_customer_flow(cn, gw,
                                   suffix=suffix,
                                   network=(ipv4_subnets[i],
                                            ipv6_subnets[i]),
                                   autostart_client=autostart_client)



import netns
from threading import Thread

from mininet.log import info, warn

from ..wifi.komondor_config import KomondorNodeConfig
from ..wifi.channel import Channel
from ..link import WifiLink
from ..config_api import ConfigAPI
from ..proxy import NamespaceProxy
from ..event import KomondorConfigChangeEvent, ChannelSwitchEvent, \
        Sub6GhzEmulatorRegistrationEvent, \
        Sub6GhzEmulatorCancelRegistrationEvent

from .router import TerranetRouter

class WifiNode(TerranetRouter):
    def __init__(self, name,
                 komondor_args={'x': 0, 'y': 0, 'z': 0},
                 coordinates={},
                 *args, **kwargs):
        komondor_args = self._insert_coorinates(coordinates,
                                                komondor_args)
        self._komondor_config = KomondorNodeConfig(name, **komondor_args)
        self.has_changed = False
        super().__init__(name,
                         *args, **kwargs)

    @property
    def komondor_config(self):
        return self._komondor_config

    def _insert_coorinates(self, coord, d={}):
        coord = { k: v for k, v in coord.items() if k in ('x', 'y', 'z') }
        d.update(coord)
        return d

    def channel_config(self):
        kcfg = self.komondor_config
        return {"primary_channel": kcfg["primary_channel"],
                "min_channel_allowed": kcfg["min_channel_allowed"],
                "max_channel_allowed": kcfg["max_channel_allowed"],
                "central_freq": kcfg["central_freq"]}

    def update_komondor_config(self, config_dict):
        self._komondor_config.update(config_dict)
        evt = KomondorConfigChangeEvent(self, config_dict)
        self.notify_sub6ghz_emulator(evt)
        return (evt.result, evt.message)

    def register_sub6ghz_emulator(self, emulator):
        self.sub6ghz_emulator = emulator
        evt = Sub6GhzEmulatorRegistrationEvent(self)
        self.notify_sub6ghz_emulator(evt)
        return (evt.result, evt.message)

    def unregister_sub6ghz_emulator(self):
        self.sub6ghz_emulator = None
        evt = Sub6GhzEmulatorCancelRegistrationEvent(self)
        self.notify_sub6ghz_emulator(evt)
        return (evt.result, evt.message)

    def clear_changed(self):
        self.has_changed = False

    def notify_sub6ghz_emulator(self, evt, wait=True):
        self.has_changed = True
        if self.sub6ghz_emulator:
            self.sub6ghz_emulator.update(evt)
        if wait:
            evt.wait()
        self.clear_changed()


class WifiAccessPoint(WifiNode):
    def __init__(self, name, ssid,
                 available_channels=[Channel(32)],
                 primary_channel=None,
                 sub6ghz_emulator=None,
                 komondor_args={},
                 *args, **kwargs):
        komondor_args.update({"type": 0})
        self.available_channels = available_channels
        channel_config = available_channels[0].komondor_channel_params
        if not primary_channel:
            primary_channel = channel_config["min_channel_allowed"]
        komondor_args.update({"wlan_code": ssid,
                              "primary_channel": primary_channel})
        komondor_args.update(channel_config)
        self.sub6ghz_emulator = sub6ghz_emulator
        super().__init__(name,
                         komondor_args=komondor_args,
                         *args, **kwargs)

    def connected_stations(self):
        stations = []
        for intf in self.intfList():
            link = intf.link
            if link and isinstance(link, WifiLink):
                node1, node2 = link.intf1.node, link.intf2.node
                if node1 == self and isinstance(node2, WifiStation):
                    stations.append(node2)
                elif node2 == self and isinstance(node1, WifiStation):
                    stations.append(node1)
        return stations

    def get_channel_config(self):
        return self.channel_config()


class ConfigurableWifiAccessPoint(WifiAccessPoint):
    def __init__(self, name, ssid,
                 proxy_port=None,
                 *args, **kwargs):
        super().__init__(name, ssid,
                         *args, **kwargs)
        self.run_config_api_thread()
        self.proxy = None
        self.proxy_port = proxy_port
        if proxy_port:
            self.proxy = self.run_namespace_proxy()

    def run_config_api_thread(self):
        nspid = self.pid
        with netns.NetNS(nspid=nspid):
            api = ConfigAPI(self.name, self)
            kwargs = {"host": "::",
                      "port": 80}
            api_thread = Thread(target=api.run, kwargs=kwargs)
            api_thread.daemon = True
            api_thread.start()

    def run_namespace_proxy(self):
        nspid = self.pid
        proxy = NamespaceProxy(nspid, self.proxy_port)
        info('Starting Namespace proxy for node {}'
             'with NSPID {} on port {}.\n'.format(self.name,
                                                  proxy.nspid,
                                                  proxy.port))
        proxy.start_proxy_process()
        info('Proxy started at process {}.\n'.format(proxy.process.pid))
        return proxy

    def switch_channel(self, channel, primary_channel=None):
        channel_params = channel.komondor_channel_params
        if not primary_channel:
            primary_channel = channel_params["min_channel_allowed"]
        min_channel_allowed = channel_params["min_channel_allowed"]
        max_channel_allowed = channel_params["max_channel_allowed"]
        central_freq = channel_params["central_freq"]
        old_channel_cfg = self.get_channel_config()
        new_channel_cfg = {"primary_channel": primary_channel,
                           "min_channel_allowed": min_channel_allowed,
                           "max_channel_allowed": max_channel_allowed,
                           "central_freq": central_freq}
        self.update_komondor_config(new_channel_cfg)
        evt = ChannelSwitchEvent(self,
                                 old_channel_cfg,
                                 new_channel_cfg)
        self.notify_sub6ghz_emulator(evt)
        return (evt.result, evt.message)

    def terminate(self):
        if self.proxy and self.proxy.process:
            try:
                pid = self.proxy.process.pid
                info('Terminating proxy process with pid {}.\n'.format(pid))
                self.proxy.process.terminate()
            except Exception as e:
                warn('Could not kill proxy process: \n'
                     '{}\n'.format(e))
        super().terminate()


class WifiStation(WifiNode):
    def __init__(self,
                 name,
                 komondor_args={},
                 *args, **kwargs):
        komondor_args.update({"type": 1})
        super().__init__(name,
                         komondor_args=komondor_args,
                         *args, **kwargs)



from .iperf import IperfHost, IperfClient, IperfReverseClient, IperfServer
from .router import TerranetRouter
from .terranode import ClientNode, DistributionNode, DistributionNode60, \
                       DistributionNode5_60, Gateway
from .wifi import WifiNode, WifiAccessPoint, ConfigurableWifiAccessPoint, \
                  WifiStation


__all__ = [
        'IperfHost', 'IperfClient', 'IperfReverseClient', 'IperfServer',
        'TerranetRouter',
        'ClientNode', 'DistributionNode60', 'DistributionNode5_60',
        'Gateway',
        'WifiNode', 'WifiAccessPoint', 'ConfigurableWifiAccessPoint',
        'WifiStation'
]



from .terratopo import Terratopo
from .virtual_fiber import HybridVirtualFiberTopo
from .backup import HybridBackupTerragraphTopo


__all__ = [
    'Terratopo',
    'HybridVirtualFiberTopo',
    'HybridBackupTerragraphTopo'
]



from mininet.node import OVSSwitch
from ipmininet.ipnet import IPNet
from ipmininet.ipnet import IPLink, IPIntf
from ipmininet.router.config import OpenrRouterConfig
from .link import TerraLink, TerraIntf
from .node import TerranetRouter, ClientNode, DistributionNode, \
                  DistributionNode60, DistributionNode5_60, \
                  IperfHost, IperfClient, IperfServer, \
                  WifiNode, WifiAccessPoint, WifiStation
from .wifi.sub6ghz_emulator import Sub6GhzEmulator
from .wifi.komondor_config import KomondorSystemConfig
from .influx.customerstats import CustomerstatsContinuousQueries
from .influx.switchstats import SwitchstatsContinuousQueries


class Terranet(IPNet):
    def __init__(self,
                 topo=None,
                 komondor_system_cfg=None,
                 sub6ghz_emulator=None,
                 komondor_config_dir=None,
                 router=DistributionNode60,
                 config=OpenrRouterConfig,
                 link=IPLink,
                 intf=IPIntf,
                 switch=OVSSwitch,
                 ipBase=u'10.0.0.0/9',
                 ip6Base=u'fd00:0:0::0/49',
                 max_v6_prefixlen=96,
                 *args, **kwargs):
        if not komondor_config_dir:
            if topo:
                komondor_config_dir = topo.komondor_config_dir()
        if not sub6ghz_emulator:
            sub6ghz_emulator = Sub6GhzEmulator(
                net=self,
                komondor_config_dir=komondor_config_dir)
        self.sub6ghz_emulator = sub6ghz_emulator
        super().__init__(topo=topo,
                         router=router,
                         config=config,
                         link=link,
                         intf=intf,
                         switch=switch,
                         *args, **kwargs)

    def build(self):
        super().build()
        for node in self.wifi_nodes():
            node.register_sub6ghz_emulator(self.sub6ghz_emulator)
        if not self.sub6ghz_emulator.build_komondor():
            self.sub6ghz_emulator = None

        if self.sub6ghz_emulator:
            self.sub6ghz_emulator.apply_wifi_config()

        CustomerstatsContinuousQueries.drop_cqs()
        SwitchstatsContinuousQueries.drop_cqs()
        CustomerstatsContinuousQueries.create_cqs()
        SwitchstatsContinuousQueries.create_cqs()

    def start(self):
        super().start()
        self.start_iperf_hosts()

    def stop(self):
        super().stop()
        CustomerstatsContinuousQueries.drop_cqs()
        SwitchstatsContinuousQueries.drop_cqs()

    def start_iperf_hosts(self):
        # resolve iperf server addresses
        iperf_server_names = [x.name for x in self.get_iperf_servers()]
        for iperf_client in self.get_iperf_clients():
            if iperf_client.host in iperf_server_names:
                iperf_server = self[iperf_client.host]
                iperf_server_ip = iperf_server.intfList()[0].ip6
                iperf_client.host = iperf_server_ip

        # autostart iperf host if enabled
        for iperf_host in self.get_iperf_hosts():
            if iperf_host.autostart:
                if iperf_host.autostart_params:
                    iperf_host.run_iperf(iperf_host.autostart_params)
                else:
                    iperf_host.run_iperf()

    def terranet_routers(self):
        return list(filter(lambda x: isinstance(x, TerranetRouter),
                           self.routers))

    def client_nodes(self):
        return list(filter(lambda x: isinstance(x, ClientNode),
                           self.terranet_routers()))

    def distribution_nodes(self):
        return [x for x in self.terranet_routers()
                    if isinstance(x, DistributionNode)]

    def distribution_nodes_60(self):
        return list(filter(lambda x: isinstance(x, DistributionNode60),
                           self.terranet_routers()))

    def distribution_nodes_5_60(self):
        return list(filter(lambda x: isinstance(x, DistributionNode5_60),
                           self.terranet_routers()))

    def wifi_nodes(self):
        return list(filter(lambda x: isinstance(x, WifiNode),
                           self.terranet_routers()))

    def access_points(self):
        return list(filter(lambda x: isinstance(x, WifiAccessPoint),
                           self.terranet_routers()))

    def stations(self):
        return list(filter(lambda x: isinstance(x, WifiStation),
                           self.terranet_routers()))

    def get_nodes_by_komondor_setting(self, key, value):
        return list(filter(lambda x: x.komondor_config[key] == value,
                           self.terranet_routers()))

    def connected_wifi_nodes(self, distribution_node_5_60):
        connectionsTo()
        wlan_code = distribution_node_5_60.komondor_config["wlan_code"]
        nodes_with_wlan_code = self.get_nodes_by_komondor_setting(
                                        "wlan_code", wlan_code)
        return list(filter(lambda x: isinstance(x, ClientNode),
                           nodes_with_wlan_code))

    def get_iperf_hosts(self):
        return list(filter(lambda x: isinstance(x, IperfHost),
                           self.hosts))

    def get_iperf_clients(self):
        return list(filter(lambda x: isinstance(x, IperfClient),
                           self.hosts))

    def get_iperf_servers(self):
        return list(filter(lambda x: isinstance(x, IperfServer),
                           self.hosts))



from threading import Event


class TerranetEvent:
    def __init__(self, cls=Event):
        self._event = cls()
        self.result = None
        self.message = None

    @property
    def event(self):
        return self._event

    def is_set(self):
        return self._event.is_set()

    isSet = is_set

    def set(self):
        self._event.set()

    def clear(self):
        self._event.clear()

    def wait(self, timeout=None):
        self._event.wait(timeout=timeout)


class KomondorConfigChangeEvent(TerranetEvent):
    def __init__(self, node, update):
        self.node = node
        self.update = update
        super().__init__()


class ChannelSwitchEvent(TerranetEvent):
    def __init__(self, node,
                 old_channel_config,
                 new_channel_config):
        self.node = node
        self.old_channel_config = old_channel_config
        self.new_channel_config = new_channel_config
        super().__init__()


class Sub6GhzEmulatorRegistrationEvent(TerranetEvent):
    def __init__(self, node):
        self.node = node
        super().__init__()


class Sub6GhzEmulatorCancelRegistrationEvent(TerranetEvent):
    def __init__(self, node):
        self.node = node
        super().__init__()



import functools

from influxdb import InfluxDBClient

from ryu.base import app_manager
from ryu.controller.handler import set_ev_cls
from ryu.lib import hub

from .customer_flow_pipeline import EventCustomerFlowAdded, \
    EventCustomerFlowRemoved
from .ipv6_address_helper import IPv6AddressHelper

class CustomerAllocationMonitor(app_manager.RyuApp):

    INTERVAL = 10
    INFLUX_DATABASE = 'switchstats'
    INFLUX_USERNAME = 'admin'
    INFLUX_PASSWORD = 'admin'

    def __init__(self, *args, **kwars):
        super().__init__(*args, **kwars)

        self.customer_allocation = {}
        self.influx = InfluxDBClient(username=self.__class__.INFLUX_USERNAME,
                                     password=self.__class__.INFLUX_PASSWORD,
                                     database=self.__class__.INFLUX_DATABASE)
        self.monitor_thread = hub.spawn(self._monitor)

    def customers(self):
        return functools.reduce(lambda x, y: [*x, *y['customers']],
                                self.customer_allocation.values(),
                                [])

    def customer_count(self):
        return len(self.customers())

    def dn_customers(self, dn_id):
        return self.customer_allocation[dn_id]['customers']

    def dn_customer_count(self, dn_id):
        return len(self.dn_customers(dn_id))

    def _monitor(self):
        while True:
            self._influx_write_total_customer_count()
            self._influx_write_dn_customer_counts()
            hub.sleep(self.__class__.INTERVAL)

    def _influx_write_total_customer_count(self):
        datapoints = [
                {
                    'measurement': 'total_customer_count',
                    'fields': {
                           'value': self.customer_count()
                    }
                }
        ]
        self.influx.write_points(datapoints)
        return datapoints

    def _influx_write_dn_customer_counts(self):
        datapoints = []

        for dn_id in self.customer_allocation.keys():
            datapoint = {
                'measurement': 'customer_counts',
                'tags': {
                    'DN': dn_id
                },
                'fields': {
                    'value': self.dn_customer_count(dn_id)
                }
            }
            datapoints.append(datapoint)

        self.influx.write_points(datapoints)
        return datapoints

    @set_ev_cls(EventCustomerFlowAdded)
    def _customer_flow_added_handler(self, ev):
        self.logger.debug('CustomerAllocationMonitor: Received event '
                          '{}.'.format(ev.__class__.__name__))
        ipv6 = ev.customer_ipv6
        dn_id = IPv6AddressHelper.distribution_id(ipv6)
        if not self.customer_allocation.get(dn_id):
            self.customer_allocation[dn_id] = {'customers': []}
        self.customer_allocation[dn_id]['customers'].append(ipv6)
        self.logger.info('CustomerAllocationMonitor: Allocation between '
                         'customer {} and DN {} added.'.format(ipv6, dn_id))
        return self.customer_allocation

    @set_ev_cls(EventCustomerFlowRemoved)
    def _customer_flow_removed_handler(self, ev):
        self.logger.debug('CustomerAllocationMonitor: Received event '
                          '{}.'.format(ev.__class__.__name__))
        ipv6 = ev.customer_ipv6
        dn_id = IPv6AddressHelper.distribution_id(ipv6)
        self.customer_allocation[dn_id]['customers'].remove(ipv6)
        self.logger.info('CustomerAllocationMonitor: Allocation between '
                         'customer {} and DN {} removed.'.format(ipv6, dn_id))
        return self.customer_allocation



import os
import functools

import requests

from ryu import cfg
import oslo_config
from ryu.controller.handler import set_ev_cls
from ryu.lib import hub

from terranet.wifi.komondor_config import KomondorConfig, \
                                          read_komondor_configs, \
                                          read_komondor_results

from .customer_flow_pipeline import EventCustomerFlowAdded, \
    EventCustomerFlowRemoved
from .customer_allocation_monitor import CustomerAllocationMonitor
from .ipv6_address_helper import IPv6AddressHelper


opts = (cfg.StrOpt('topo', default='HybridVirtualFiberTopo',
                   help='Topo file used to locate Komondor cache'),
        oslo_config.cfg.DictOpt('ssids', default={},
                                help='DN id to WiFi SSID mapping'),
        oslo_config.cfg.DictOpt('proxy_ports', default={},
                                help='DN id to NamespaceProxy port mapping')
        )
cfg.CONF.register_opts(opts, 'channel_oracle')


class ChannelAssignmentOracle(CustomerAllocationMonitor):
    def __init__(self, *args, **kwars):
        super().__init__(*args, **kwars)

        self.CONF = self.CONF['channel_oracle']
        self.logger.info('ChannelAssignmentOracle: Configured ssids {}.'
                         .format(self.CONF.ssids))
        self.logger.info('ChannelAssignmentOracle: '
                         'Configured proxy ports {}.'
                         .format(self.CONF.proxy_ports))
        topo_dir = os.path.normpath(
            os.path.join(os.path.dirname(__file__),
            '../../topo')
        )
        topo = self.CONF.topo
        komondor_config_dir = f'{topo_dir}/.komondor/{topo}'
        self.logger.info('ChannelAssignmentOracle: Using komondor cache {}.'
                         .format(komondor_config_dir))

        self.komondor_config_dir = os.path.abspath(komondor_config_dir)
        self.komondor_input_dir = f'{self.komondor_config_dir}/input'
        self.komondor_output_dir = f'{self.komondor_config_dir}/output'
        self.komondor_configs, self.komondor_results = \
            self._read_komondor_cache()

    def _read_komondor_cache(self):
        configs = read_komondor_configs(self.komondor_input_dir)
        results = read_komondor_results(self.komondor_output_dir)
        return (configs, results)

    @set_ev_cls(EventCustomerFlowAdded)
    def _customer_flow_added_handler(self, ev):
        super()._customer_flow_added_handler(ev)
        channel_configs = self._channel_configurations()
        if channel_configs:
            self.assign_channels(channel_configs)
        return self.customer_allocation

    @set_ev_cls(EventCustomerFlowRemoved)
    def _customer_flow_removed_handler(self, ev):
        super()._customer_flow_removed_handler(ev)
        channel_configs = self._channel_configurations()
        if channel_configs:
            self.assign_channels(channel_configs)
        return self.customer_allocation

    def assign_channels(self, channel_configs):
        for dn_id, conf in channel_configs.items():
            channel = conf['channel']
            port = int(self.CONF.proxy_ports[str(dn_id)])
            self._request_channel_switch(channel, port)

    def _request_channel_switch(self, channel, port,
                                host='http://localhost'):
        url = f'{host}:{port}/channel'
        payload = {'channel': channel}
        self.logger.info('ChannelAssignmentOracle: Posting channel switch '
                         'url {} payload {}.'.format(url, payload))
        r = requests.post(url, json=payload)
        self.logger.info('ChannelAssignmentOracle: Channel switch response code '
                         '{}, message {}.'.format(r.status_code, r.text))
        return r

    def _channel_configurations(self):
        oracle_dict = {}
        if len(self.CONF.ssids) > len(self.customer_allocation):
            self.logger.warn('ChannelAssignmentOracle: Not all DNs have been '
                             'added to the database yet. Skipping channel '
                             'assignment.')
            return None
        res = sorted(self._compute_oracle_dict(),
                     key=lambda x: x['jainXtpt'],
                     reverse=True)
        best = res[0]
        self.logger.info('ChannelAssignmentOracle: Best configuration {} for '
                         'current customer distribution {}.'
                         .format(best, self.customer_allocation))

        komondor_config = KomondorConfig(cfg_file=best['config_file'])
        dn_channel_configs = komondor_config.ap_channel_configurations()
        for config in dn_channel_configs:
            ssid = config['wlan_code']
            dn_id = self.get_dn_id_by_ssid(ssid)
            channel = config['channel']
            oracle_dict[dn_id] = {'channel': channel}
        return oracle_dict

    def _compute_oracle_dict(self):
        metrics_list = []
        for file, result in self.komondor_results.items():
            result_file = os.path.join(self.komondor_output_dir, file)
            config_file = os.path.join(self.komondor_input_dir, file)
            result_metrics = self._analyse_komondor_result(result)
            result_metrics['result_file'] = result_file
            result_metrics['config_file'] = config_file
            metrics_list.append(result_metrics)
        return metrics_list

    def _analyse_komondor_result(self, komondor_result):
        configured_ssids = sorted(
            [x for x in self.CONF.ssids.values()])
        ssids = komondor_result.wlans()
        assert configured_ssids == ssids

        total_throughput = komondor_result.total_throughput()
        wlan_throughputs = self._extract_wlan_throughputs(komondor_result)
        throughputs = functools.reduce(
            lambda x, y: [*x, *y['customer_throughputs']],
            wlan_throughputs.values(), [])
        fairness_index = self._calculate_jains_fairness_index(throughputs)

        metrics = {
            'total_throughput': total_throughput,
            'fairness_index': fairness_index,
            'jainXtpt': fairness_index * total_throughput,
            'wlan_throughputs': wlan_throughputs
        }

        return metrics

    def _extract_wlan_throughputs(self, komondor_result):
        ssids = komondor_result.wlans()
        throughputs = {}
        for ssid in ssids:
            throughputs[ssid] = {}
            throughput = komondor_result.wlan_throughput(ssid)
            throughputs[ssid]['throughput'] = throughput
            dn_id = self.get_dn_id_by_ssid(ssid)
            customer_count = self.dn_customer_count(dn_id)
            total_throughput = komondor_result.wlan_throughput(ssid)
            throughput_per_customer = (total_throughput / customer_count)
            customer_throughputs = \
                [throughput_per_customer for _ in range(0, customer_count)]
            throughputs[ssid]['customer_throughputs'] = customer_throughputs
        return throughputs

    def _calculate_jains_fairness_index(self, throughputs=[]):
        customer_count = len(throughputs)
        sum_throughputs_squared = pow(sum(throughputs), 2)
        squared_throughputs = [pow(x, 2) for x in throughputs]
        sum_squared_throughputs = sum(squared_throughputs)
        return (sum_throughputs_squared /
            (customer_count * sum_squared_throughputs))


    def get_dn_id_by_ssid(self, ssid):
        for dn_id in self.customer_allocation.keys():
            ssid_conf = self.CONF.ssids[str(dn_id)]
            if ssid_conf == ssid:
                return dn_id
        return None



import os
import itertools
import collections
from configparser import ConfigParser

from .channel import Channel


def _read_komondor_files(directory, cls):
    config_dict = collections.OrderedDict()
    files = [f for f in sorted(os.listdir(directory)) if f.endswith(".cfg")]
    for cfg_file in files:
        path = os.path.join(directory, cfg_file)
        cfg = cls(path)
        config_dict[cfg_file] = cfg
    return config_dict


def read_komondor_configs(dir):
    return _read_komondor_files(dir, KomondorConfig)


def read_komondor_results(dir):
    return _read_komondor_files(dir, KomondorResult)


class KomondorBaseConfig(ConfigParser):
    def __init__(self,
                 cfg_file=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.cfg_file = cfg_file
        if self.cfg_file:
            self.read(cfg_file)

    # Keep case for options https://stackoverflow.com/a/23007845
    def optionxform(self, optionstr):
        return optionstr

    def sections_by_value(self, key, value):
        return [self[x] for x in self.sections() if self[x].get(key) == value]

    def nodes(self):
        return [self[x] for x in self.sections() if not x == 'System']


class KomondorConfig(KomondorBaseConfig):
    def __init__(self, cfg_file=None):
        super(KomondorConfig, self).__init__(cfg_file=cfg_file)

    def system(self):
        return self["System"]

    def access_points(self):
        return self.sections_by_value("type", "0")

    def stations(self):
        return self.sections_by_value("type", "1")

    def nodes_by_wlan_code(self, wlan_code):
        return self.sections_by_value("wlan_code", wlan_code)

    def get_stations_by_access_point(self, ap):
        return list(filter(lambda x: x["type"] == "1",
                           self.nodes_by_wlan_code(ap["wlan_code"])))

    def wifi5_links_for_access_point(self, ap):
        return [(ap.name, sta.name)
                for sta in self.get_stations_by_access_point(ap)]

    def wifi5GHz_links(self):
        return reduce(lambda x, y: x + self.wifi5_links_for_access_point(y),
                      self.access_points(), [])

    def wifi60GHz_links(self):
        aps = map(lambda x: x.name,
                  self.sort_west_to_east(self.access_points()))
        return [(aps[i], aps[i+1]) for i, _ in enumerate(aps)
                if i+1 < len(aps)]

    def sort_west_to_east(self, nodes):
        return sorted(nodes, key=lambda x: x["x"])

    def ap_channel_configurations(self):
        channel_config = []
        for ap in self.access_points():
            wlan_code = ap['wlan_code']
            min_channel_allowed = ap['min_channel_allowed']
            max_channel_allowed = ap['max_channel_allowed']
            ch = Channel.channel_num(min_channel_allowed, max_channel_allowed)
            channel_config.append({'name': ap.name,
                                   'wlan_code': wlan_code,
                                   'channel': ch,
                                   'min_channel_allowed': min_channel_allowed,
                                   'max_channel_allowed': max_channel_allowed})
        return channel_config



class KomondorConfigSection(collections.OrderedDict):
    def __init__(self, name, *args, **kwargs):
        self.name = name
        super().__init__(*args, **kwargs)


class KomondorSystemConfig(KomondorConfigSection):
    def __init__(self, *args, **kwargs):
        name = "System"
        defaults = collections.OrderedDict([
            ("num_channels", 50),
            ("basic_channel_bandwidth", 20),
            ("pdf_backoff", 0),
            ("pdf_tx_time", 1),
            ("packet_length", 12000),
            ("num_packets_aggregated", 64),
            ("path_loss_model_default", 5),
            ("path_loss_model_indoor_indoor", 5),
            ("path_loss_model_indoor_outdoor", 8),
            ("path_loss_model_outdoor_outdoor", 7),
            ("capture_effect", 20),
            ("noise_level", -95),
            ("adjacent_channel_model", 0),
            ("collisions_model", 0),
            ("constant_PER", 0),
            ("traffic_model", 99),
            ("backoff_type", 0),
            ("cw_adaptation", 1),
            ("pifs_activated", 0),
            ("capture_effect_model", 1)])
        config = defaults
        config.update(kwargs)
        super().__init__(name, *args, **config)


class KomondorNodeConfig(KomondorConfigSection):
    def __init__(self, name, *args, **kwargs):
        defaults = collections.OrderedDict([
            ("type", 0),
            ("wlan_code", "A"),
            ("destination_id", -1),
            ("x", 0),
            ("y", 0),
            ("z", 0),
            ("primary_channel", 0),
            ("min_channel_allowed", 0),
            ("max_channel_allowed", 0),
            ("cw", 16),
            ("cw_stage", 5),
            ("tpc_min", 30),
            ("tpc_default", 30),
            ("tpc_max", 30),
            ("cca_min", -82),
            ("cca_default", -82),
            ("cca_max", -82),
            ("tx_antenna_gain", 0),
            ("rx_antenna_gain", 0),
            ("channel_bonding_model", 1),
            ("modulation_default", 0),
            ("central_freq", 5),
            ("lambda", 10000),
            ("ieee_protocol", 1),
            ("traffic_load", 1000),
            ("node_env", "outdoor")])
        config = defaults
        config.update(kwargs)
        if config["type"] == 0:
            name = "Node_AP_{}".format(name)
        else:
            name = "Node_STA_{}".format(name)
        super().__init__(name, *args, **config)


class KomondorResult(KomondorBaseConfig):
    def __init__(self, cfg_file=None):
        super().__init__(cfg_file=cfg_file)

    def nodes_by_wlan(self, wlan):
        return self.sections_by_value('wlan', wlan)

    def total_throughput(self):
        return sum([int(self[x]['throughput']) for x in self.sections()])

    def wlan_throughput(self, wlan):
        return sum([int(x['throughput']) for x in self.nodes_by_wlan(wlan)])

    def wlans(self):
        return list(sorted(set([self[x]['wlan'] for x in self.sections()])))



from .terratopo import Terratopo
from ..wifi.channel import Channel

class HybridBackupTerragraphTopo(Terratopo):
    '''
    mmWave Distribution Network backhaul
    ====================================

            +-----+            +-----+           +-----+
            |cn_a1|            |cn_b1|           |cn_c1|
            +-----+            +-----+           +-----+
      5GHz .   |             .    |            .    |
          .                 .                 .
    +----+     |      +----+      |     +----+      |
    |ap_a|   60GHz    |ap_b|    60GHz   |ap_c|    60GHz
    +----+     |      +----+      |     +----+      |
          \                 \                 \ 
     cable \   |             \    |            \    |
            +------+  60GHz   +------+  60GHz   +------+
            | dn_a |- - - - - | dn_b |- - - - - | dn_c |
            +------+          +------+          +------+
                |
                | cable
                |
            + -_-_- +
            |  _X_  | gw
            + - - - +
              | | |
              | | | cable
             .-~~~-~.
     .- ~ ~-(         )_ _
    /                      ~ -.
   |  s1a1 s1b1 s1c1 s2c1 s3c1  \ 
    \                        _.~'
      ~- . ______________ . -



    WiFi backup edge
    ================

             h1a1              h1b1         h1c1 h2c1 h3c1    (CPEs)
               |                 |            \    |    /
       ^       |                 |             \   |   /     -- cable
       |    +-----+           +-----+           +-----+
    20 |    |cn_a1|           |cn_b1|           |cn_b2|     (STAs/ CNs)
       |    +-----+           +-----+           +-----+
       |       .                 .                 .
       |       .                 .                 .
    10 |       . CH:42           .      CH:58      .        ... 5GHz WiFi
       |       .                 .                 .
       |       .                 .                 .
       |    +------+          +------+          +------+
     0 |    | ap_a |          | ap_b |          | ap_c |    (APs/ DNs)
       |    +------+          +------+          +------+
       ----------------------------------------------->
     -15       0       25       50       75        100
    '''
    def build(self, *args, **kwargs):
        # Segment A
        channel_list_a = [
            Channel(42), # 80MHz
            Channel(38), Channel(46), # 40MHz
            Channel(36), Channel(40), Channel(44), Channel(48) # 20MHz
        ]
        dn_a = self.add_distribution_node_60('dn_a')
        ap_a = self.add_smart_access_point(
            'ap_a',
            ssid='A',
            available_channels=channel_list_a,
            coordinates={'x': 0, 'y': 0}
        )
        self.add_ip_link(dn_a, ap_a)

        cn_a1 = self.add_client_node('cn_a1', coordinates={'x': 0, 'y': 20})
        self.add_terragraph_link(dn_a, cn_a1)
        self.add_wifi_link(ap_a, cn_a1)


        # Segment B
        channel_list_b = [
            Channel(58), # 80MHz
            Channel(54), Channel(62), # 40MHz
            Channel(52), Channel(56), Channel(60), Channel(64) # 20MHz
        ]
        dn_b = self.add_distribution_node_60('dn_b')
        ap_b = self.add_smart_access_point(
            'ap_b',
            ssid='B',
            available_channels=channel_list_b,
            coordinates={'x': 50, 'y': 0}
        )
        self.add_ip_link(dn_b, ap_b)

        cn_b1 = self.add_client_node('cn_b1', coordinates={'x': 50, 'y': 20})
        self.add_terragraph_link(dn_b, cn_b1)
        self.add_wifi_link(ap_b, cn_b1)


        # Segment C
        channel_list_c = [
            Channel(58), # 80MHz
            Channel(54), Channel(62), # 40MHz
            Channel(52), Channel(56), Channel(60), Channel(64) # 20MHz
        ]
        dn_c = self.add_distribution_node_60('dn_c')
        ap_c = self.add_smart_access_point(
            'ap_c',
            ssid='C',
            available_channels=channel_list_c,
            coordinates={'x': 100, 'y': 0}
        )
        self.add_ip_link(dn_c, ap_c)

        cn_c1 = self.add_client_node('cn_c1', coordinates={'x': 100, 'y': 20})
        self.add_terragraph_link(dn_c, cn_c1)
        self.add_wifi_link(ap_c, cn_c1)


        # Add WiFi 60GHz Links between DN_A and DN_B
        self.add_terragraph_link(dn_a, dn_b)
        self.add_terragraph_link(dn_b, dn_c)

        # Add GW switch behind DN_A
        gw = self.add_gateway('s1')
        self.add_ip_link(gw, dn_a)

        # cn_a1 customer flows
        h1a1, s1a1 = self.add_customer_flow(cn_a1, gw,
            suffix='1a1',
            network=('10.145.129.0/24',
                     'fd00:0:0:8101:8001::0/80'),
            autostart_client=True
        )

        # cn_b1 customer flows
        h1b1, s1b1 = self.add_customer_flow(cn_b1, gw,
            suffix='1b1',
            network=('10.161.129.0/24',
                     'fd00:0:0:8201:8001::0/80'),
            autostart_client=True
        )

        # cn_c1 customer flows
        h1c1, s1c1 = self.add_customer_flow(cn_c1, gw,
            suffix='1c1',
            network=('10.177.129.0/24',
                     'fd00:0:0:8301:8001::0/80'),
            autostart_client=True
        )
        h2c1, s2c1 = self.add_customer_flow(cn_c1, gw,
            suffix='2c1',
            network=('10.177.130.0/24',
                     'fd00:0:0:8301:8002::0/80'),
            autostart_client=True
        )
        h3c1, s3c1 = self.add_customer_flow(cn_c1, gw,
            suffix='3c1',
            network=('10.177.131.0/24',
                     'fd00:0:0:8301:8003::0/80'),
            autostart_client=True
        )

        super().build(*args, **kwargs)



from ipmininet.router.config import OpenrRouterConfig
from ipmininet.node_description import OpenrRouterDescription


class TerranetRouterDescription(OpenrRouterDescription):
    def __new__(cls, value, *args, **kwargs):
        return super(TerranetRouterDescription, cls).__new__(cls, value,
                                                             *args, **kwargs)

    def __init__(self, o, topo):
        self.topo = topo
        super(TerranetRouterDescription, self).__init__(o, topo)

    def addDaemon(self, daemon, default_cfg_class=OpenrRouterConfig,
                  **daemon_params):
        self.topo.addDaemon(self, daemon, default_cfg_class=default_cfg_class,
                            **daemon_params)



from influxdb import InfluxDBClient


class SwitchstatsContinuousQueries:

    DATABASE = 'switchstats'
    USERNAME = 'admin'
    PASSWORD = 'admin'
    DEFAULT_INTERVAL = 10
    CONTINUOUS_QUERIES = ['dl_throughputs',
                          'ul_throughputs',
                          'sum_dl_throughputs',
                          'sum_ul_throughputs']

    @classmethod
    def create_cqs(cls,
                   username=USERNAME,
                   password=PASSWORD,
                   database=DATABASE,
                   interval=DEFAULT_INTERVAL):
        influx = InfluxDBClient(username=username,
                                password=password,
                                database=database)

        cqs = []
        cqs.append(cls.create_dlthroughputs_cq(influx, interval=interval,
                                               database=database))
        cqs.append(cls.create_ulthroughputs_cq(influx, interval=interval,
                                               database=database))
        cqs.append(cls.create_sum_dlthroughputs_cq(influx, interval=interval,
                                                   database=database))
        cqs.append(cls.create_sum_ulthroughputs_cq(influx, interval=interval,
                                                   database=database))
        return cqs

    @classmethod
    def drop_cqs(cls,
                 username=USERNAME,
                 password=PASSWORD,
                 database=DATABASE,
                 interval=DEFAULT_INTERVAL,
                 cqs=CONTINUOUS_QUERIES):
        influx = InfluxDBClient(username=username,
                                password=password,
                                database=database)
        return [ influx.drop_continuous_query(cq, database=database) \
                 for cq in cqs ]

    @classmethod
    def create_dlthroughputs_cq(cls, client,
                                cq_name='dl_throughputs',
                                interval=DEFAULT_INTERVAL,
                                database=DATABASE):
        select = (
            'SELECT non_negative_derivative(max("value"), 1s) * 8 '
            'AS "value" '
            'INTO "dl_throughputs" '
            'FROM "byte_count_dl" '
            'WHERE time > now() - {interval}s '
            'GROUP BY time({interval}s), "host"'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_ulthroughputs_cq(cls, client,
                                cq_name='ul_throughputs',
                                interval=DEFAULT_INTERVAL,
                                database=DATABASE):
        select = (
            'SELECT non_negative_derivative(max("value"), 1s) * 8 '
            'AS "value" '
            'INTO "ul_throughputs" '
            'FROM "byte_count_ul" '
            'WHERE time > now() - {interval}s '
            'GROUP BY time({interval}s), "host"'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_sum_dlthroughputs_cq(cls, client,
                                    cq_name='sum_dl_throughputs',
                                    interval=DEFAULT_INTERVAL,
                                    database=DATABASE):
        select = (
            'SELECT sum("value") '
            'AS "value" '
            'INTO "sum_dl_throughputs" '
            'FROM "dl_throughputs" '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_sum_ulthroughputs_cq(cls, client,
                                    cq_name='sum_ul_throughputs',
                                    interval=DEFAULT_INTERVAL,
                                    database=DATABASE):
        select = (
            'SELECT sum("value") '
            'AS "value" '
            'INTO "sum_ul_throughputs" '
            'FROM "ul_throughputs" '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)



import time

from mininet.log import info

from terranet.topo import HybridBackupTerragraphTopo
from terranet.net import Terranet
from terranet.controller import RyuManager
from terranet.simulator import TerranetSimulator


if __name__ == '__main__':
    topo = HybridBackupTerragraphTopo()
    net = Terranet(topo=topo,
                   build=False)
    module_path = 'terranet.ryu.app'
    apps = ('customer_flow_pipeline',
            'customer_stats_monitor',
            'customer_allocation_monitor')
    modules = [ '.'.join([module_path, x]) for x in apps ]
    ryu_manager = RyuManager('ryu-manager', modules=modules)
    net.addController(ryu_manager)
    with TerranetSimulator('handover_simulation', net, clean_after=False) as net:
        # DN_A <=> CN_A1
        dn_a = net['dn_a']
        cn_a1 = net['cn_a1']
        link_dn_a_cn_a1 = net.linksBetween(dn_a, cn_a1)[0]
        # DN_B <=> DN_B1
        dn_b = net['dn_b']
        cn_b1 = net['cn_b1']
        link_dn_b_cn_b1 = net.linksBetween(dn_b, cn_b1)[0]
        # DN_C <=> DN_C1
        dn_c = net['dn_c']
        cn_c1 = net['cn_c1']
        link_dn_c_cn_c1 = net.linksBetween(dn_c, cn_c1)[0]
        time.sleep(180)

        info('{}: Simulating link failure DN_A <=> CN_A1\n'
             .format(TerranetSimulator.time_ns()))
        link_dn_a_cn_a1.intf2.ifconfig('down')
        time.sleep(180)

        info('{}: Simulating link failure DN_B <=> CN_B1 '
             'and DN_C <=> CN_C1\n'
             .format(TerranetSimulator.time_ns()))
        link_dn_b_cn_b1.intf2.ifconfig('down')
        link_dn_c_cn_c1.intf2.ifconfig('down')
        time.sleep(180)

        info('{}: Simulating link failure recovery DN_A <=> CN_A1\n'
             .format(TerranetSimulator.time_ns()))
        link_dn_a_cn_a1.intf2.ifconfig('up')
        time.sleep(180)



from influxdb import InfluxDBClient


class CustomerstatsContinuousQueries:

    DATABASE = 'customerstats'
    USERNAME = 'admin'
    PASSWORD = 'admin'
    DEFAULT_INTERVAL = 10
    CONTINUOUS_QUERIES = ['rx_throughputs',
                          'sum_rx_throughputs',
                          'dn1_rx_throughputs',
                          'dn2_rx_throughputs',
                          'dn3_rx_throughputs',
                          'janes_customer_count',
                          'janes_pow2_sum_throughputs',
                          'janes_sum_pow2_throughputs',
                          'janes_fairness_index']

    @classmethod
    def create_cqs(cls,
                   username=USERNAME,
                   password=PASSWORD,
                   database=DATABASE,
                   interval=DEFAULT_INTERVAL):
        influx = InfluxDBClient(username=username,
                                password=password,
                                database=database)

        cqs = []
        cqs.append(cls.create_rxthroughputs_cq(influx, interval=interval,
                                               database=database))
        cqs.append(cls.create_sum_rxthroughputs_cq(influx, interval=interval,
                                                   database=database))
        for dn_subnet_prefix in [81, 82, 83]:
            cqs.append(cls.create_dn_rxthroughputs_cq(
                influx,
                interval=interval,
                dn_subnet_prefix=dn_subnet_prefix,
                database=database))
        cqs.append(cls.create_customer_count_cq(influx, interval=interval,
                                                database=database))
        cqs.append(cls.create_pow2_sum_throughputs_cq(influx,
                                                      interval=interval,
                                                      database=database))
        cqs.append(cls.create_sum_pow2_throughputs_cq(influx,
                                                      interval=interval,
                                                      database=database))
        cqs.append(cls.create_fairness_index_cq(influx, interval=interval,
                                                database=database))
        return cqs

    @classmethod
    def drop_cqs(cls,
                 username=USERNAME,
                 password=PASSWORD,
                 database=DATABASE,
                 interval=DEFAULT_INTERVAL,
                 cqs=CONTINUOUS_QUERIES):
        influx = InfluxDBClient(username=username,
                                password=password,
                                database=database)
        return [ influx.drop_continuous_query(cq, database=database) \
                 for cq in cqs ]

    @classmethod
    def create_rxthroughputs_cq(cls, client,
                                cq_name='rx_throughputs',
                                interval=DEFAULT_INTERVAL,
                                database=DATABASE):
        select = (
            'SELECT non_negative_derivative(max("value"), 1s) * 8 '
            'AS "value" '
            'INTO "rx_throughputs" '
            'FROM "rx_bytes_value" '
            'WHERE time > now() - {interval}s '
            'GROUP BY time({interval}s), "host"'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_sum_rxthroughputs_cq(cls, client,
                                    cq_name='sum_rx_throughputs',
                                    interval=DEFAULT_INTERVAL,
                                    database=DATABASE):
        select = (
            'SELECT sum("value") '
            'AS "value" '
            'INTO "sum_rx_throughputs" '
            'FROM "rx_throughputs" '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_dn_rxthroughputs_cq(cls, client,
                                    cq_name=None,
                                    dn_subnet_prefix=81,
                                    interval=DEFAULT_INTERVAL,
                                    database=DATABASE):
        cqs = []
        dn_id = int('0x{}'.format(dn_subnet_prefix), 16) - 0x80
        if not cq_name:
            cq_name = 'dn{}_rx_throughputs'.format(dn_id)
        select = (
            'SELECT sum("value") '
            'AS "dn{dn_id}" '
            'INTO "dn_rx_throughputs" '
            'FROM "rx_throughputs" '
            'WHERE "host" =~ /fd00::{subnet_prefix}.*/'
            'GROUP BY time({interval}s)'
        ).format(dn_id=dn_id,
                 subnet_prefix=dn_subnet_prefix,
                 interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_customer_count_cq(cls, client,
                                 cq_name='janes_customer_count',
                                 interval=DEFAULT_INTERVAL,
                                 database=DATABASE):
        select = (
            'SELECT count("last") '
            'AS "customer_count" '
            'INTO "janes_fairness_index" '
            'FROM (SELECT last("value")'
                  'FROM "rx_throughputs"'
                  'GROUP BY "host") '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_pow2_sum_throughputs_cq(cls, client,
                                       cq_name='janes_pow2_sum_throughputs',
                                       interval=DEFAULT_INTERVAL,
                                       database=DATABASE):
        select = (
            'SELECT pow(last("value"), 2) '
            'AS "pow2_sum_throughputs" '
            'INTO "janes_fairness_index" '
            'FROM "sum_rx_throughputs" '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_sum_pow2_throughputs_cq(cls, client,
                                          cq_name='janes_sum_pow2_throughputs',
                                          interval=DEFAULT_INTERVAL,
                                          database=DATABASE):
        select = (
            'SELECT sum(pow2_throughputs) '
            'AS "sum_pow2_throughputs" '
            'INTO "janes_fairness_index" '
            'FROM (SELECT pow(last("value"), 2)'
                  'AS "pow2_throughputs"'
                  'FROM "rx_throughputs"'
                  'GROUP BY "host") '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)

    @classmethod
    def create_fairness_index_cq(cls, client,
                                 cq_name='janes_fairness_index',
                                 interval=DEFAULT_INTERVAL,
                                 database=DATABASE):
        select = (
            'SELECT last("index") '
            'AS "value" '
            'INTO "janes_fairness_index" '
            'FROM (SELECT "pow2_sum_throughputs"'
                          '/ "customer_count"'
                          '/ "sum_pow2_throughputs"'
                          'AS "index"'
                          'FROM "janes_fairness_index") '
            'GROUP BY time({interval}s)'
        ).format(interval=interval)
        return client.create_continuous_query(cq_name, select,
                                              database=database)



from influxdb import InfluxDBClient

from ryu.base import app_manager
from ryu.controller import ofp_event
from ryu.controller.handler import MAIN_DISPATCHER, DEAD_DISPATCHER
from ryu.controller.handler import set_ev_cls
from ryu.lib import hub
from ryu.controller import dpset


class CustomerStatsMonitor(app_manager.RyuApp):
    '''
    https://osrg.github.io/ryu-book/en/html/traffic_monitor.html
    '''

    INTERVAL = 5
    INFLUX_DATABASE = 'switchstats'
    INFLUX_USERNAME = 'admin'
    INFLUX_PASSWORD = 'admin'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.datapaths = {}
        self.table_ids = []
        self.influx = InfluxDBClient(username=self.__class__.INFLUX_USERNAME,
                                     password=self.__class__.INFLUX_PASSWORD,
                                     database=self.__class__.INFLUX_DATABASE)
        self.monitor_thread = hub.spawn(self._monitor)

    @set_ev_cls(ofp_event.EventOFPStateChange,
                [MAIN_DISPATCHER, DEAD_DISPATCHER])
    def _state_change_handler(self, ev):
        datapath = ev.datapath
        if ev.state == MAIN_DISPATCHER:
            if datapath.id not in self.datapaths:
                self.logger.info('CustomerStatsMonitor: Register '
                                 'datapath {}.'.format(datapath.id))
                self.datapaths[datapath.id] = datapath
        elif ev.state == DEAD_DISPATCHER:
            if datapath.id in self.datapaths:
                self.logger.info('CustomerStatsMonitor: Unregister '
                                 'datapath {}.'.format(datapath.id))
                del self.datapaths[datapath.id]

    def _monitor(self):
        while True:
            for dp in self.datapaths.values():
                self._request_table_stats(dp)
                self._request_customer_stats(dp)
            hub.sleep(self.__class__.INTERVAL)

    def _request_table_stats(self, datapath):
        self.logger.debug('CustomerStatsMonitor: Sending TableStatsRequest to '
                          'datapath {}.'.format(datapath.id))
        ofp_parser = datapath.ofproto_parser

        req = ofp_parser.OFPTableStatsRequest(datapath, 0)
        datapath.send_msg(req)

    @set_ev_cls(ofp_event.EventOFPTableStatsReply, MAIN_DISPATCHER)
    def _table_stats_reply_handler(self, ev):
        self.logger.debug('CustomerStatsMonitor: TableStatsReply received.')
        for stat in ev.msg.body:
            table_id = stat.table_id
            if (not self._is_customer_table(table_id) or
                not stat.active_count):
                continue
            if not table_id in self.table_ids:
                self.table_ids.append(table_id)
                self.logger.info('CustomerStatsMonitor: Added id {} to table_ids.'
                                  .format(table_id))

    def _is_customer_table(self, table_id):
        return (table_id > 3 and
                table_id < 100)

    def _request_customer_stats(self, datapath):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        for table_id in self.table_ids:
            req = parser.OFPFlowStatsRequest(datapath, table_id=table_id)
            datapath.send_msg(req)
            self.logger.debug('CustomerStatsMonitor: '
                              'FlowStatsRequest send for table_id {} '
                              'to datapath {}.'
                              .format(table_id, datapath.id))

    @set_ev_cls(ofp_event.EventOFPFlowStatsReply, MAIN_DISPATCHER)
    def _flow_stats_reply_handler(self, ev):
        self.logger.debug('CustomerStatsMonitor: FlowStatsReply received.')
        body = ev.msg.body

        for stat in [flow for flow in body if flow.priority > 1]:
            self.logger.debug('CustomerStatsMonitor: Received stats: '
                              'table_id {}, '
                              'match {}, '
                              'packet_count {}, '
                              'byte_count {}.'.format(stat.table_id,
                                                      stat.match,
                                                      stat.packet_count,
                                                      stat.byte_count))

            if stat.match.get('ipv6_src'):
                ipv6 = stat.match['ipv6_src']
                measurements = ('packet_count_ul', 'byte_count_ul')

            if stat.match.get('ipv6_dst'):
                ipv6 = stat.match['ipv6_dst']
                measurements = ('packet_count_dl', 'byte_count_dl')

            if measurements:
                datapoints = [
                    {
                        'measurement': measurements[0],
                        'tags': {
                            'host': ipv6
                        },
                        'fields': {
                            'value': stat.packet_count
                        }
                    },
                    {
                        'measurement': measurements[1],
                        'tags': {
                            'host': ipv6
                        },
                        'fields': {
                            'value': stat.byte_count
                        }
                    }
                ]

                self.influx.write_points(datapoints)



import os
from setuptools import setup, find_packages

# https://packaging.python.org/guides/single-sourcing-package-version/
# 4th option
with open(os.path.join(os.getcwd(), 'VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    name='terranet',
    version=version,
    packages=find_packages(),
    author='butjar',
    author_email='',
    description='Terranet: mmWave Distribution Network Emulator',
    keywords='terranet terragraph wifi wifi6 60GHz mmWave',
    install_requires=[
        'collectd',
        'configparser',
        'Flask',
        'influxdb',
        'netns',
        'numpy',
        'proxy.py',
        'requests',
        'ryu',
        'whichcraft',
    ],
    include_package_data=True,
    package_data={
        'terranet': ['topo/.komondor/**/*',
                     'ryu/app/cfg/*.conf']
    },
    scripts=['terranet/examples/run_virtual_fiber_net.py',
             'terranet/examples/run_hybrid_backup_net.py',
             'terranet/simulation/simulate_hybrid_handover.py',
             'terranet/simulation/simulate_customer_arrivals.py',
             'terranet/simulation/simulate_channel_oracle.py'],
)



import os
from mininet.node import Controller

from .ryu import app


class RyuManager(Controller):
    def __init__(self, name,
                 *ryuArgs,
                 modules=['terranet.ryu.app.mac_learning_pipeline'],
                 **kwargs):
        self.modules = modules
        command = 'ryu-manager'
        ryuArgs = ' '.join(ryuArgs)
        cargs = '--ofp-tcp-listen-port %s {ryuArgs} {modules}' \
                .format(ryuArgs=ryuArgs,
                        modules=' '.join(self.modules))
        cdir = os.path.dirname(app.__file__)
        kwargs.update({'command': command,
                       'cargs': cargs,
                       'cdir': cdir})
        super().__init__(name, **kwargs)



from ryu.base import app_manager
from ryu.controller import ofp_event
from ryu.controller.handler import CONFIG_DISPATCHER, MAIN_DISPATCHER
from ryu.controller.handler import set_ev_cls
from ryu.ofproto import ofproto_v1_3
from ryu.lib.packet import packet
from ryu.lib.packet import ethernet
from ryu.lib.packet import ipv6


class MacLearningPipeline(app_manager.RyuApp):
    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]
    SRC_MAC_TABLE = 0
    DST_MAC_TABLE = 1

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)
    def switch_features_handler(self, ev):
        datapath = ev.msg.datapath
        self._init_src_mac_table(datapath)
        self._init_dst_mac_table(datapath)

    def _init_src_mac_table(self, datapath):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        table_id = self.__class__.SRC_MAC_TABLE
        actions = [parser.OFPActionOutput(ofproto.OFPP_CONTROLLER)]
        inst = [parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS,
                                             actions)]
        mod = parser.OFPFlowMod(datapath=datapath,
                                table_id=table_id,
                                priority=0,
                                match=parser.OFPMatch(),
                                instructions=inst)
        datapath.send_msg(mod)
        self.logger.info("MacLearningPipeline: Initialized src MAC table {}. "
                         "On table-miss send to controller.".format(table_id))

    def _init_dst_mac_table(self, datapath):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        table_id = self.__class__.DST_MAC_TABLE
        actions = [parser.OFPActionOutput(ofproto.OFPP_FLOOD, 0)]
        inst = [parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS,
                                             actions)]
        mod = parser.OFPFlowMod(datapath=datapath,
                                table_id=table_id,
                                priority=0,
                                match=parser.OFPMatch(),
                                instructions=inst)
        datapath.send_msg(mod)
        self.logger.info("MacLearningPipeline: Initialized dst MAC table {}. "
                         "On table-miss flood.".format(table_id))

    @set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)
    def _packet_in_handler(self, ev):
        msg = ev.msg
        in_port = msg.match['in_port']
        table_id = msg.table_id
        self.logger.info("MacLearningPipeline: Packet in at port {} "
                         "from table {}.".format(in_port,
                                                 table_id))

        if table_id == self.__class__.SRC_MAC_TABLE:
            self._handle_new_src_mac(ev)

    def _handle_new_src_mac(self, ev):
        msg = ev.msg
        datapath = msg.datapath
        parser = datapath.ofproto_parser
        buffer_id = msg.buffer_id
        in_port = msg.match['in_port']
        pkt = packet.Packet(msg.data)
        eth = pkt.get_protocols(ethernet.ethernet)[0]
        self.logger.info("MacLearningPipeline: New src MAC {} "
                         "at port {}.".format(eth.src,
                                              in_port))

        self._add_flow_mod_mac_dst(datapath, eth.src, in_port)
        self._add_flow_mod_mac_src(datapath, eth.src, in_port, buffer_id)

    def _add_flow_mod_mac_dst(self, datapath, eth_dst, out_port):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        table_id = self.__class__.DST_MAC_TABLE
        match = parser.OFPMatch(eth_dst=eth_dst)
        actions = [parser.OFPActionOutput(out_port)]
        inst = [parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS,
                                             actions)]
        mod = parser.OFPFlowMod(datapath=datapath,
                                table_id=table_id,
                                match=match,
                                instructions=inst)
        datapath.send_msg(mod)
        self.logger.info("MacLearningPipeline: "
                         "Sending packets with dst MAC {} "
                         "out at port {}. "
                         "Flow mod added to table {}.".format(eth_dst,
                                                              out_port,
                                                              table_id))

    def _add_flow_mod_mac_src(self, datapath, eth_src, in_port, buffer_id):
        parser = datapath.ofproto_parser
        table_id = self.__class__.SRC_MAC_TABLE
        match = parser.OFPMatch(in_port=in_port,
                                eth_src=eth_src)
        dst_table_id = self.__class__.DST_MAC_TABLE
        inst = [parser.OFPInstructionGotoTable(dst_table_id)]
        mod = parser.OFPFlowMod(datapath=datapath,
                                table_id=table_id,
                                buffer_id=buffer_id,
                                match=match,
                                instructions=inst)
        datapath.send_msg(mod)
        self.logger.info("MacLearningPipeline: Process packets from port {} "
                         "with src MAC {} "
                         "at table {}. "
                         "Flow mod added to table {}.".format(in_port,
                                                              eth_src,
                                                              dst_table_id,
                                                              table_id))



class Channel:
    channel_map = {
        # 2.402 - 2.494 GHz
        1:   {"f0": 2412,
              "width":  20,
              "komondor": {"min_channel_allowed":  1,
                           "max_channel_allowed":  1,
                           "central_freq":         2.412}},
        2:   {"f0": 2417,
              "width":  20,
              "komondor": {"min_channel_allowed":  2,
                           "max_channel_allowed":  2,
                           "central_freq":         2.417}},
        3:   {"f0": 2422,
              "width":  20,
              "komondor": {"min_channel_allowed":  3,
                           "max_channel_allowed":  3,
                           "central_freq":         2.422}},
        4:   {"f0": 2427,
              "width":  20,
              "komondor": {"min_channel_allowed":  4,
                           "max_channel_allowed":  4,
                           "central_freq":         2.427}},
        5:   {"f0": 2432,
              "width":  20,
              "komondor": {"min_channel_allowed":  5,
                           "max_channel_allowed":  5,
                           "central_freq":         2.432}},
        6:   {"f0": 2437,
              "width":  20,
              "komondor": {"min_channel_allowed":  6,
                           "max_channel_allowed":  6,
                           "central_freq":         2.437}},
        7:   {"f0": 2442,
              "width":  20,
              "komondor": {"min_channel_allowed":  7,
                           "max_channel_allowed":  7,
                           "central_freq":         2.442}},
        8:   {"f0": 2447,
              "width":  20,
              "komondor": {"min_channel_allowed":  8,
                           "max_channel_allowed":  8,
                           "central_freq":         2.447}},
        9:   {"f0": 2452,
              "width":  20,
              "komondor": {"min_channel_allowed":  9,
                           "max_channel_allowed":  9,
                           "central_freq":         2.452}},
        10:  {"f0": 2457,
              "width":  20,
              "komondor": {"min_channel_allowed": 10,
                           "max_channel_allowed": 10,
                           "central_freq":         2.457}},
        11:  {"f0": 2462,
              "width":  20,
              "komondor": {"min_channel_allowed": 11,
                           "max_channel_allowed": 11,
                           "central_freq":         2.462}},
        12:  {"f0": 2467,
              "width":  20,
              "komondor": {"min_channel_allowed": 12,
                           "max_channel_allowed": 12,
                           "central_freq":         2.467}},
        13:  {"f0": 2472,
              "width":  20,
              "komondor": {"min_channel_allowed": 13,
                           "max_channel_allowed": 13,
                           "central_freq":         2.472}},
        14:  {"f0": 2484,
              "width":  20,
              "komondor": {"min_channel_allowed": 14,
                           "max_channel_allowed": 14,
                           "central_freq":         2.484}},
        # 5.150 - 5.330 GHz
        32:  {"f0": 5160,
              "width":  20,
              "komondor": {"min_channel_allowed": 15,
                           "max_channel_allowed": 15,
                           "central_freq":         5.160}},
        34:  {"f0": 5170,
              "width":  40,
              "komondor": {"min_channel_allowed": 15,
                           "max_channel_allowed": 16,
                           "central_freq":         5.170}},
        36:  {"f0": 5180,
              "width":  20,
              "komondor": {"min_channel_allowed": 16,
                           "max_channel_allowed": 16,
                           "central_freq":         5.180}},
        38:  {"f0": 5190,
              "width":  40,
              "komondor": {"min_channel_allowed": 16,
                           "max_channel_allowed": 17,
                           "central_freq":         5.190}},
        40:  {"f0": 5200,
              "width":  20,
              "komondor": {"min_channel_allowed": 17,
                           "max_channel_allowed": 17,
                           "central_freq":         5.200}},
        42:  {"f0": 5210,
              "width":  80,
              "komondor": {"min_channel_allowed": 16,
                           "max_channel_allowed": 19,
                           "central_freq":         5.210}},
        44:  {"f0": 5220,
              "width":  20,
              "komondor": {"min_channel_allowed": 18,
                           "max_channel_allowed": 18,
                           "central_freq":         5.220}},
        46:  {"f0": 5230,
              "width":  40,
              "komondor": {"min_channel_allowed": 18,
                           "max_channel_allowed": 19,
                           "central_freq":         5.230}},
        48:  {"f0": 5240,
              "width":  20,
              "komondor": {"min_channel_allowed": 19,
                           "max_channel_allowed": 19,
                           "central_freq":         5.240}},
        50:  {"f0": 5250,
              "width": 160,
              "komondor": {"min_channel_allowed": 16,
                           "max_channel_allowed": 23,
                           "central_freq":         5.250}},
        52:  {"f0": 5260,
              "width":  20,
              "komondor": {"min_channel_allowed": 20,
                           "max_channel_allowed": 20,
                           "central_freq":         5.260}},
        54:  {"f0": 5270,
              "width":  40,
              "komondor": {"min_channel_allowed": 20,
                           "max_channel_allowed": 21,
                           "central_freq":         5.270}},
        56:  {"f0": 5280,
              "width":  20,
              "komondor": {"min_channel_allowed": 21,
                           "max_channel_allowed": 21,
                           "central_freq":         5.280}},
        58:  {"f0": 5290,
              "width":  80,
              "komondor": {"min_channel_allowed": 20,
                           "max_channel_allowed": 23,
                           "central_freq":         5.290}},
        60:  {"f0": 5300,
              "width":  20,
              "komondor": {"min_channel_allowed": 22,
                           "max_channel_allowed": 22,
                           "central_freq":         5.300}},
        62:  {"f0": 5310,
              "width":  40,
              "komondor": {"min_channel_allowed": 22,
                           "max_channel_allowed": 23,
                           "central_freq":         5.310}},
        64:  {"f0": 5320,
              "width":  20,
              "komondor": {"min_channel_allowed": 23,
                           "max_channel_allowed": 23,
                           "central_freq":         5.320}},
        # 5.330 - 5350 GHz
        68:  {"f0": 5340,
              "width":  20,
              "komondor": {"min_channel_allowed": 24,
                           "max_channel_allowed": 24,
                           "central_freq":         5.340}},
        # 5.470 - 5.490 GHz
        96:  {"f0": 5480,
              "width":  20,
              "komondor": {"min_channel_allowed": 25,
                           "max_channel_allowed": 25,
                           "central_freq":         5.480}},
        # 5490 - 5650 GHz
        100: {"f0": 5500,
              "width":  20,
              "komondor": {"min_channel_allowed": 26,
                           "max_channel_allowed": 26,
                           "central_freq":         5.500}},
        102: {"f0": 5510,
              "width":  40,
              "komondor": {"min_channel_allowed": 26,
                           "max_channel_allowed": 27,
                           "central_freq":         5510}},
        104: {"f0": 5520,
              "width":  20,
              "komondor": {"min_channel_allowed": 27,
                           "max_channel_allowed": 27,
                           "central_freq":         5.520}},
        106: {"f0": 5530,
              "width":  80,
              "komondor": {"min_channel_allowed": 26,
                           "max_channel_allowed": 29,
                           "central_freq":         5.530}},
        108: {"f0": 5540,
              "width":  20,
              "komondor": {"min_channel_allowed": 28,
                           "max_channel_allowed": 28,
                           "central_freq":         5.540}},
        110: {"f0": 5550,
              "width":  40,
              "komondor": {"min_channel_allowed": 28,
                           "max_channel_allowed": 29,
                           "central_freq":         5.550}},
        112: {"f0": 5560,
              "width":  20,
              "komondor": {"min_channel_allowed": 29,
                           "max_channel_allowed": 29,
                           "central_freq":         5.560}},
        114: {"f0": 5570,
              "width": 160,
              "komondor": {"min_channel_allowed": 26,
                           "max_channel_allowed": 33,
                           "central_freq":         5.570}},
        116: {"f0": 5580,
              "width":  20,
              "komondor": {"min_channel_allowed": 30,
                           "max_channel_allowed": 30,
                           "central_freq":         5.580}},
        118: {"f0": 5590,
              "width":  40,
              "komondor": {"min_channel_allowed": 30,
                           "max_channel_allowed": 31,
                           "central_freq":         5.590}},
        120: {"f0": 5600,
              "width":  20,
              "komondor": {"min_channel_allowed": 31,
                           "max_channel_allowed": 31,
                           "central_freq":         5.600}},
        122: {"f0": 5610,
              "width":  80,
              "komondor": {"min_channel_allowed": 30,
                           "max_channel_allowed": 33,
                           "central_freq":         5.610}},
        124: {"f0": 5620,
              "width":  20,
              "komondor": {"min_channel_allowed": 32,
                           "max_channel_allowed": 32,
                           "central_freq":         5.620}},
        126: {"f0": 5630,
              "width":  40,
              "komondor": {"min_channel_allowed": 32,
                           "max_channel_allowed": 33,
                           "central_freq":         5.630}},
        128: {"f0": 5640,
              "width":  20,
              "komondor": {"min_channel_allowed": 33,
                           "max_channel_allowed": 33,
                           "central_freq":         5.640}},
        # 5.650 - 5.730GHz
        132: {"f0": 5660,
              "width":  20,
              "komondor": {"min_channel_allowed": 34,
                           "max_channel_allowed": 34,
                           "central_freq":         5.660}},
        134: {"f0": 5670,
              "width":  40,
              "komondor": {"min_channel_allowed": 34,
                           "max_channel_allowed": 35,
                           "central_freq":         5.670}},
        136: {"f0": 5680,
              "width":  20,
              "komondor": {"min_channel_allowed": 35,
                           "max_channel_allowed": 35,
                           "central_freq":         5.680}},
        138: {"f0": 5690,
              "width":  80,
              "komondor": {"min_channel_allowed": 34,
                           "max_channel_allowed": 37,
                           "central_freq":         5.690}},
        140: {"f0": 5700,
              "width":  20,
              "komondor": {"min_channel_allowed": 36,
                           "max_channel_allowed": 36,
                           "central_freq":         5.700}},
        142: {"f0": 5710,
              "width":  40,
              "komondor": {"min_channel_allowed": 36,
                           "max_channel_allowed": 37,
                           "central_freq":         5.710}},
        144: {"f0": 5720,
              "width":  20,
              "komondor": {"min_channel_allowed": 37,
                           "max_channel_allowed": 37,
                           "central_freq":         5.720}},
        # 5.735 - 5.815 GHz
        149: {"f0": 5745,
              "width":  20,
              "komondor": {"min_channel_allowed": 38,
                           "max_channel_allowed": 38,
                           "central_freq":         5.745}},
        151: {"f0": 5755,
              "width":  40,
              "komondor": {"min_channel_allowed": 38,
                           "max_channel_allowed": 39,
                           "central_freq":         5.755}},
        153: {"f0": 5765,
              "width":  20,
              "komondor": {"min_channel_allowed": 39,
                           "max_channel_allowed": 39,
                           "central_freq":         5.765}},
        155: {"f0": 5775,
              "width":  80,
              "komondor": {"min_channel_allowed": 38,
                           "max_channel_allowed": 41,
                           "central_freq":         5.775}},
        157: {"f0": 5785,
              "width":  20,
              "komondor": {"min_channel_allowed": 40,
                           "max_channel_allowed": 40,
                           "central_freq":         5.785}},
        159: {"f0": 5795,
              "width":  40,
              "komondor": {"min_channel_allowed": 40,
                           "max_channel_allowed": 41,
                           "central_freq":         5.795}},
        161: {"f0": 5805,
              "width":  20,
              "komondor": {"min_channel_allowed": 41,
                           "max_channel_allowed": 41,
                           "central_freq":         5.805}},
        # 5.835 - 5.855 GHz
        169: {"f0": 5845,
              "width":  20,
              "komondor": {"min_channel_allowed": 42,
                           "max_channel_allowed": 42,
                           "central_freq":         5.845}},
        # 5.855 - 5.875 GHz
        173: {"f0": 5865,
              "width":  20,
              "komondor": {"min_channel_allowed": 43,
                           "max_channel_allowed": 43,
                           "central_freq":         5.865}}}

    @classmethod
    def channel_num(cls, min_channel_allowed, max_channel_allowed):
        for ch, val in cls.channel_map.items():
            k_conf = val['komondor']
            if k_conf['min_channel_allowed'] == int(min_channel_allowed) and \
               k_conf['max_channel_allowed'] == int(max_channel_allowed):
                   return ch
        return None

    def __init__(self, channel_num):
        self.num = channel_num
        channel_dict = Channel.channel_map[channel_num]
        self.f0 = channel_dict["f0"]
        self.width = channel_dict["width"]
        self.komondor_channel_params = channel_dict["komondor"]

    def __eq__(self, other):
        return self.num == other.num

    def __lt__(self, other):
        return self.num < other.num



import os
import uuid
import itertools
from ipaddress import IPv4Network, IPv6Network

from ipmininet.link import IPLink
from ipmininet.iptopo import IPTopo
from ipmininet.router.config import OpenrRouterConfig

from ..router_config import TerranetRouterDescription
from ..node import TerranetRouter, ClientNode, \
                   DistributionNode60, DistributionNode5_60, \
                   WifiAccessPoint, ConfigurableWifiAccessPoint, \
                   Gateway, \
                   IperfReverseClient, IperfServer
from ..link import WifiLink, TerragraphLink
from ..wifi.komondor_config import KomondorSystemConfig


class Terratopo(IPTopo):
    def __init__(self,
                 komondor_system_config=None,
                 create_komondor_dirs=True,
                 *args, **kwargs):
        if not komondor_system_config:
            self.komondor_system_config = KomondorSystemConfig()
        self.create_komondor_dirs = create_komondor_dirs
        super().__init__(*args, **kwargs)

    def build(self, *args, **kwargs):
        if self.create_komondor_dirs:
            self._create_komondor_dirs()

        super().build(*args, **kwargs)


    def komondor_config_dir(self):
        topo_dir = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(topo_dir, '.komondor',
                            str(self.__class__.__name__))

    def _create_komondor_dirs(self):
        komondor_input_dir = os.path.join(self.komondor_config_dir(),
                                          'input')
        komondor_output_dir = os.path.join(self.komondor_config_dir(),
                                          'output')
        if not os.path.isdir(self.komondor_config_dir()):
            os.makedirs(self.komondor_config_dir(), exist_ok=True)

        if not os.path.isdir(komondor_input_dir):
            os.makedirs(komondor_input_dir, exist_ok=True)

        if not os.path.isdir(komondor_output_dir):
            os.makedirs(komondor_output_dir, exist_ok=True)

    def addRouter(self, name,
                  cls=TerranetRouter,
                  routerDescription=TerranetRouterDescription,
                  config=OpenrRouterConfig,
                  **kwargs):
        return routerDescription(self.addNode(str(name),
                                              isRouter=True,
                                              cls=cls,
                                              **kwargs), self)

    def add_client_node(self, name, **opts):
        return self.addRouter(name, is_client_node=True, cls=ClientNode,
                              **opts)

    def add_distribution_node_60(self, name, **opts):
        return self.addRouter(name, is_distribution_node=True,
                              cls=DistributionNode60, **opts)

    def add_distribution_node_5_60(self, name, **opts):
        return self.addRouter(name, is_distribution_node=True,
                              is_access_point=True,
                              cls=DistributionNode5_60, **opts)

    def add_access_point(self, name, **opts):
        return self.addRouter(name, is_access_point=True,
                              cls=WifiAccessPoint, **opts)

    def add_smart_access_point(self, name, **opts):
        return self.addRouter(name, is_access_point=True,
                              cls=ConfigurableWifiAccessPoint, **opts)

    def add_gateway(self, name, **opts):
        return self.addSwitch(name, cls=Gateway, **opts)

    def add_iperf_reverse_client(self, name, host=None, **opts):
        return self.addHost(name, cls=IperfReverseClient,
                            host=host, **opts)

    def add_iperf_server(self, name, **opts):
        return self.addHost(name, cls=IperfServer, **opts)

    def add_ip_link(self, node1, node2, *args, **kwargs):
        return self.addLink(node1, node2, cls=IPLink)

    def add_wifi_link(self, node1, node2, *args, **kwargs):
        return self.addLink(node1, node2, cls=WifiLink)

    def add_terragraph_link(self, node1, node2):
        return self.addLink(node1, node2, cls=TerragraphLink)

    def is_client_node(self, node):
        return self.isNodeType(node, 'is_client_node')

    def is_distribution_node(self, node):
        return self.isNodeType(node, 'is_distribution_node')

    def client_nodes(self, sort=True):
        return list(filter(self.is_client_node,
                           self.nodes(sort)))

    def distribution_nodes(self, sort=True):
        return list(filter(self.is_distribution_node,
                           self.nodes(sort)))

    def terranet_routers(self, sort=True):
        return client_nodes(sort=sort) + distribution_nodes(sort=sort)

    def _get_customer_ip(self, ip):
        cn_v4_net = IPv4Network(ip[0].decode('utf-8'), strict=False)
        customer_v4_net = next(
            itertools.islice(cn_v4_net.subnets(), 1, None))
        customer_v4_ip = "{netaddr}/{prefix}".format(
                netaddr=str(customer_v4_net.network_address),
                prefix=cn_v4_net.prefixlen)

        cn_v6_net = IPv6Network(ip[1].decode('utf-8'), strict=False)
        customer_v6_net = next(
            itertools.islice(cn_v6_net.subnets(), 1, None))
        customer_v6_ip = "{netaddr}/{prefix}".format(
                netaddr=str(customer_v6_net.network_address),
                prefix=cn_v6_net.prefixlen)
        return (customer_v4_ip, customer_v6_ip)

    def _get_ip(self, network,
                 cn_prefix4=24,
                 cpe_prefix4=16,
                 cn_prefix6=80,
                 cpe_prefix6=64):
        net4 = IPv4Network(network[0])
        cn4 =  "{}/{}".format(net4[1], cn_prefix4)
        cpe4 = "{}/{}".format(net4[2], cpe_prefix4)

        net6 = IPv6Network(network[1])
        cn6 = "{}/{}".format(net6[1], cn_prefix6)
        cpe6 = "{}/{}".format(net6[2], cpe_prefix6)
        return ((cn4, cn6), (cpe4, cpe6))

    def add_customer_flow(self,
                          cn,
                          gateway,
                          suffix=None,
                          client_prefix="h",
                          server_prefix="s",
                          network=("10.0.0.0/24",
                                   "fd00:0:0:8000:8000::0/80"),
                          autostart_client=False,
                          autostart_server=True,
                          *args, **kwargs):
        if not suffix:
            suffix_len = 10 - max(len(client_prefix), len(server_prefix))
            suffix = uuid.uuid4().hex[:suffix_len]
        cpe_name = "{}{}".format(client_prefix, suffix)
        server_name = "{}{}".format(server_prefix, suffix)
        cpe = self.add_iperf_reverse_client(cpe_name,
                                            server_name,
                                            autostart=autostart_client)
        link = self.add_wifi_link(cn, cpe)
        (cn_ip, cpe_ip) = self._get_ip(network)
        link[cn].addParams(ip=cn_ip)
        link[cpe].addParams(ip=cpe_ip)

        server = self.add_iperf_server(server_name,
                                       autostart=True)
        self.add_ip_link(gateway, server)
        return (cpe, server)



import ipaddress


class IPv6AddressHelper:
    @classmethod
    def parse_address(cls, ip):
        return ipaddress.IPv6Address(str(ip))

    @classmethod
    def prefix(cls, ip):
        ip_parts = cls.parse_address(ip).exploded.split(":")
        return "{}:{}:{}".format(ip_parts[0], ip_parts[1], ip_parts[2])

    @classmethod
    def subnet_id(cls, ip):
        ip_parts = cls.parse_address(ip).exploded.split(":")
        return ip_parts[3]

    @classmethod
    def subnet_prefix(cls, ip):
        return "{}:{}".format(cls.prefix(ip), cls.subnet_id(ip))

    @classmethod
    def interface_id(cls, ip):
        ip_parts = cls.parse_address(ip).exploded.split(":")
        return "{}:{}:{}:{}".format(ip_parts[4], ip_parts[5], ip_parts[6],
                                    ip_parts[7])

    @classmethod
    def distribution_id(cls, ip):
        subnet_id = cls.subnet_id(ip)
        distribution_id = int("0x{}".format(subnet_id[:2]), 16) - 0x80
        return distribution_id

    @classmethod
    def client_id(cls, ip):
        subnet_id = cls.subnet_id(ip)
        client_id = int("0x{}".format(subnet_id[2:]), 16)
        return client_id

    @classmethod
    def distribution_net(cls, ip):
        prefix = cls.prefix(ip)
        distribution_id = cls.distribution_id(ip)
        distribution_subnet = distribution_id + 0x80
        return ipaddress.ip_network(
                u"{}:{:x}00::0/56".format(prefix,
                                          distribution_subnet))

    @classmethod
    def client_net(cls, ip):
        subnet_prefix = cls.subnet_prefix(ip)
        return ipaddress.ip_network(u"{}::/64".format(subnet_prefix))

    @classmethod
    def customer_net(cls, ip):
        subnet_prefix = cls.subnet_prefix(ip)
        return ipaddress.ip_network(u"{}:8000::/65".format(subnet_prefix))

    @classmethod
    def is_customer_address(cls, ip):
        subnet_prefix = cls.subnet_prefix(ip)
        customer_net = cls.customer_net(ip)
        if cls.parse_address(ip) in customer_net:
            return True
        return False



import os
import inspect
import time
import shlex, subprocess
import glob
import shutil
import logging
from influxdb import InfluxDBClient
import json
import csv

from mininet.log import lg, info, warn, error
from ipmininet.clean import cleanup

class TerranetSimulator():
    def __init__(self, name, net,
                 home_dir=None,
                 clean_before=True,
                 clean_after=True,
                 log_level='info',
                 compress=True,
                 influx_config={'host': 'localhost',
                                'port': 8086,
                                'user': 'admin',
                                'password': 'admin',
                                'databases': ['customerstats',
                                              'switchstats']},
                 *args, **kwargs):
        self.name = name
        self.net = net
        if not home_dir:
            home_dir = '/var/terranet/{}'.format(self.name)
        self.home_dir = home_dir
        self.clean_before = clean_before
        self.clean_after = clean_after
        self.log_level = log_level
        self.compress = compress
        self.simulation_start = self.time_ns()
        self.simulation_end = None
        self.net_started = None
        self.net_stopped = None
        self.simulation_dir = None
        self.log_name = '{}.log'.format(
            self.__class__.__name__)
        self.simulation_dir = os.path.join(
            self.home_dir, str(self.simulation_start))
        self.logfile = os.path.join(
            self.simulation_dir, self.log_name)
        self.influx_config = influx_config
        self.archive = None

    @staticmethod
    def time_ns():
        return int(round(time.time() * 1000000000))

    def __enter__(self):
        self.setup()
        self.run()
        return self.net

    def __exit__(self, type, value, traceback):
        self.stop()
        self.teardown()
        return True

    def run(self):
        info('***** Simulation {} started at: {}\n'
             .format(self.name, self.simulation_start))
        net_name = self.net.__class__.__name__
        info('***** Starting net {}\n'.format(net_name))

        self.net.start()
        self.net_started = self.time_ns()
        info('***** {} started at {}\n'
             .format(net_name, self.net_started))

    def setup(self):
        info('***** Setup\n')
        self._setup_logfile()
        if self.clean_before:
            self.clean(caller='setup')

    def _setup_logfile(self):
        os.makedirs(self.simulation_dir, exist_ok=True)

        if self.logfile:
            lg.addHandler(logging.FileHandler(self.logfile))
            lg.setLogLevel(self.log_level)
            info('***** Logfile: {}\n'.format(self.logfile))


    def stop(self):
        net_name = self.net.__class__.__name__
        self.net_stopped = self.time_ns()
        info('***** {} stopped at: {}\n'
            .format(net_name, self.net_stopped))
        self.net.stop()
        self.simulation_end = self.time_ns()
        info('***** Simulation {} ended at: {}\n'
            .format(self.name, self.simulation_end))

    def teardown(self):
        info('***** Teardown\n')
        self._archive_topo_file()
        self._archive_logs()
        self._archive_komondor_cache()
        self._archive_influx()
        if self.compress:
            self._compress_simulation_dir()
            self._delete_simulation_dir()
        if self.clean_after:
            self.clean(caller='teardown')

    def sh(self, command,
           **kwargs):
        defaults = {'shell': True,
                    'universal_newlines': True,
                    'stdout': subprocess.PIPE,
                    'stderr': subprocess.PIPE}
        defaults.update(kwargs)
        info('\n/bin/sh# {}\n'.format(command))
        with subprocess.Popen(command, **defaults) as p:
            stdout, stderr = p.communicate()
            if stdout:
                info('/bin/sh(OUT)# {}\n'.format(stdout))
            if stderr:
                error('/bin/sh(ERR)# {}\n'.format(stderr))

    def clean(self, caller=None):
        info('***** Cleaning logs\n')
        shutil.rmtree('/var/tmp/log', ignore_errors=True)
        self._clean_influx()
        if caller == 'teardown':
            if self.archive:
                info('***** Deleting compressed simulation dir\n')
                self._delete_simulation_dir()
            info('***** (IP)mininet cleanup\n')
            cleanup()

    def _clean_influx(self):
        info('***** Cleaning influxdb\n')
        for db in self.influx_config['databases']:
            influx = InfluxDBClient(self.influx_config['host'],
                self.influx_config['port'], self.influx_config['user'],
                self.influx_config['password'], db)
            info('***** Dropping database {}\n'.format(db))
            influx.drop_database(db)
            info('***** Creating database {}\n'.format(db))
            influx.create_database(db)

    def archive_file(self, file,
                     path=None,
                     sub_dirs=None,
                     new_file_name=None):
        src = os.path.abspath(file)
        if not new_file_name:
            new_file_name = os.path.basename(file)
        if not path:
            path = self.simulation_dir
        if sub_dirs:
            path = os.path.join(path, sub_dirs)
        os.makedirs(path, exist_ok=True)
        dst = os.path.join(path, new_file_name)
        copy = shutil.copy2(src, dst)
        info('***** File {} archived to {}\n'.format(src, copy))
        return copy

    def archive_dirs(self, dir,
                     path=None,
                     sub_dirs=None,
                     new_dir_name=None):
        src = os.path.abspath(dir)
        if not new_dir_name:
            new_dir_name = os.path.basename(dir)
        if not path:
            path = self.simulation_dir
        if sub_dirs:
            path = os.path.join(path, sub_dirs)
        dst = os.path.join(path, new_dir_name)
        copy = shutil.copytree(src, dst)
        info('***** Directory tree {} archived to {}\n'.format(src, copy))

    def _archive_topo_file(self):
        topo_file = os.path.abspath(inspect.getfile(
            self.net.topo.__class__))
        info('***** Archiving topo file: {}\n'.format(topo_file))
        self.archive_file(topo_file)

    def _archive_influx(self):
        influx_archive = os.path.join(self.simulation_dir, 'influx')
        for db in self.influx_config['databases']:
            db_archive = os.path.join(influx_archive, db)
            os.makedirs(db_archive, exist_ok=True)
            influx = InfluxDBClient(self.influx_config['host'],
                self.influx_config['port'], self.influx_config['user'],
                self.influx_config['password'], db)
            measurements = [x['name'] for x in influx.get_list_measurements()]
            for m in measurements:
                info('***** Archiving measurement {} '
                     'from database: {}\n'.format(m, db))
                q = 'SELECT * FROM {measurement} ' \
                    'WHERE time > {start} ' \
                    'AND time < {end}'.format(measurement=m,
                                              start=self.net_started,
                                              end=self.net_stopped)
                rs = influx.query(q, database=db, epoch='ns')
                json_file = os.path.join(db_archive, '{}.json'.format(m))
                self._write_resultset(json_file, rs, format='json')
                dat_file = os.path.join(db_archive, '{}.dat'.format(m))
                self._write_resultset(dat_file, rs, format='csv')

    def _write_resultset(self, file, result_set, format='json'):
        if format == 'json':
            with open(file, 'w') as json_file:
                json.dump(result_set.raw, json_file, indent=4, sort_keys=True)
            return json_file
        if format == 'csv':
            with open(file, 'w', newline='') as dat_file:
                writer = csv.writer(dat_file, dialect='unix', delimiter='\t')
                # write header
                writer.writerow(result_set.raw['series'][0]['columns'])
                writer.writerows(result_set.raw['series'][0]['values'])
            return dat_file
        return None

    def _archive_logs(self):
        info('***** Archiving logs\n')
        log_archive = os.path.join(self.simulation_dir, 'log')
        os.makedirs(log_archive, exist_ok=True)
        self.archive_dirs('/var/tmp/log/', sub_dirs='log/openr')
        log_files = [*glob.glob('/tmp/*.log'),
                     *glob.glob('/var/log/iperf_*.log')]
        for log in log_files:
            self.archive_file(log, sub_dirs='log')

    def _archive_openr_configs(self):
        info('***** Archiving OpenR config files\n')
        for conf in glob.glob('/tmp/openr_*.cfg'):
            self.archive_file(conf, sub_dirs='conf/openr')

    def _archive_komondor_cache(self):
        info('***** Archiving komondor cache\n')
        cache_dir = os.path.join(self.net.topo.komondor_config_dir())
        self.archive_dirs(os.path.join(cache_dir, 'input'),
                          sub_dirs='komondor/input')
        self.archive_dirs(os.path.join(cache_dir, 'output'),
                          sub_dirs='komondor/output')

    def _compress_simulation_dir(self):
        info('***** Compressing simulation directory\n')
        root_dir = os.path.dirname(self.simulation_dir)
        base_dir = os.path.basename(self.simulation_dir)
        archive_name = os.path.join(root_dir, base_dir)
        self.archive = shutil.make_archive(archive_name,
                                           format='gztar',
                                           root_dir=root_dir,
                                           base_dir=base_dir)

    def _delete_simulation_dir(self):
        shutil.rmtree(self.simulation_dir)
